import asyncio
import attrs
import json
import logging
import pylsl as lsl
import socket
import time
import threading
import typing as tp

from . import getKeyForStreamInfo, getEquivalentKeysForStreamInfo, inferStreamKey
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.ZMQConnector import ZMQConnectorServer, ZMQConnectorClient, getNewPort


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@attrs.define()
class LSLStreamResolver:
    _resolveStreamsTimeout: float = 0.01  # in s, if this is too short, some streams may be missed

    sigStreamDetected: Signal = attrs.field(init=False, factory=lambda: Signal((str, lsl.StreamInfo)))
    sigStreamLost: Signal = attrs.field(init=False, factory=lambda: Signal((str, lsl.StreamInfo)))

    _availableStreams: tp.Dict[str, lsl.StreamInfo] = attrs.field(init=False, factory=dict)

    def __attrs_post_init__(self):
        pass

    @property
    def availableStreams(self):
        return self._availableStreams

    def updateAvailableStreams(self):
        logger.debug('Resolving available streams')
        streamInfos = lsl.resolve_streams(
            wait_time=self._resolveStreamsTimeout)  # note that if this is too short, not all streams may be returned
        # TODO: set this resolve call to run with longer wait time in a separate thread/process and just send back
        #  stream names (not sure if streamInfo objects themselves are thread-safe) to be used for establishing
        #  connections in main thread
        logger.debug('Found %d streams' % len(streamInfos))
        streamIDs = []
        if True:
            for streamInfo in streamInfos:
                streamID = streamInfo.source_id()
                if streamInfo.hostname() == socket.gethostname():
                    # if hostname matches our hostname, refer to as @localhost instead of @hostname
                    # for easier config file migrations
                    hostnameOrLocalhost = 'localhost'
                else:
                    hostnameOrLocalhost = streamInfo.hostname()

                if len(streamID) == 0:
                    # if no source_id specified, construct unique(ish) key from stream name + hostname
                    # NOTE: could add other metadata into this ID to make more likely to be unique (e.g. num chan, srate)
                    streamID = streamInfo.name() + '_' + hostnameOrLocalhost
                else:
                    streamID += '@' + hostnameOrLocalhost
                streamIDs.append(streamID)
        else:
            streamIDs = [streamInfo.source_id() for streamInfo in streamInfos]
            assert all(len(streamID) > 0 for streamID in streamIDs)  # all streams should have a source ID set

        logger.debug('StreamIDs: %s' % (streamIDs,))
        assert len(set(streamIDs)) == len(streamIDs), "all stream source IDs should be unique"
        streams = {streamID: streamInfo for streamID, streamInfo in zip(streamIDs, streamInfos)}
        for streamKey in self._availableStreams.copy():
            if streamKey not in streams:
                self._onStreamLost(streamKey)

        for streamKey, streamInfo in streams.items():
            if streamKey not in self._availableStreams:
                self._onStreamDetected(streamKey, streamInfo)

    def _streamInfoAsDict(self, streamKey: str, streamInfo: lsl.StreamInfo) -> tp.Dict[str, tp.Any]:
        return dict(
            key=streamKey,
            name=streamInfo.name(),
            hostname=streamInfo.hostname(),
            sourceID=streamInfo.source_id(),
            type=streamInfo.type()
        )

    def _onStreamDetected(self, streamKey: str, streamInfo: lsl.StreamInfo):
        logger.info('New stream detected: %s' % streamKey)
        assert streamKey not in self._availableStreams
        self._availableStreams[streamKey] = streamInfo
        self.sigStreamDetected.emit(streamKey, streamInfo)

    def _onStreamLost(self, streamKey: str):
        logger.info('Stream lost: %s' % streamKey)
        try:
            assert streamKey in self._availableStreams
        except AssertionError as e:
            raise e
        streamInfo = self._availableStreams.pop(streamKey)
        self.sigStreamLost.emit(streamKey, streamInfo)

    def markStreamAsLost(self, streamKey: str):
        self._onStreamLost(streamKey)


@attrs.define(kw_only=True)
class _ThreadedLSLStreamResolver_Daemon(LSLStreamResolver):
    pollPeriod: float = 1.
    _pubPort: int
    _repPort: int

    _pollTask: tp.Optional[asyncio.Task] = attrs.field(init=False, default=None)

    _connector: ZMQConnectorServer = attrs.field(init=False)

    def __attrs_post_init__(self):
        LSLStreamResolver.__attrs_post_init__(self)
        self._connector = ZMQConnectorServer(obj=self, pubSubPort=self._pubPort, reqRepPort=self._repPort)

        # do not start async polling here; let master start it via connector, to make sure master doesn't miss any messages

    async def _poll(self):
        while True:
            self.updateAvailableStreams()
            await asyncio.sleep(max(self.pollPeriod - self._resolveStreamsTimeout, 0))

    def startAsyncPolling(self):
        if self._pollTask is None:
            self._pollTask = asyncio.create_task(asyncTryAndLogExceptionOnError(self._poll))

    def _onStreamDetected(self, streamKey: str, streamInfo: lsl.StreamInfo):
        LSLStreamResolver._onStreamDetected(self, streamKey=streamKey, streamInfo=streamInfo)
        logger.info('Publishing streamDetected message')
        self._connector.publish([b'streamDetected', json.dumps(self._streamInfoAsDict(
            streamKey=streamKey, streamInfo=streamInfo)).encode('utf-8')])

    def _onStreamLost(self, streamKey: str):
        LSLStreamResolver._onStreamLost(self, streamKey=streamKey)
        logger.info('Publishing streamLost message')
        self._connector.publish([b'streamLost', streamKey.encode('utf-8')])

    def markStreamAsLost(self, streamKey: str):
        if streamKey not in self._availableStreams:
            return  # silently ignore if we already marked this stream as lost
        LSLStreamResolver.markStreamAsLost(self, streamKey=streamKey)

    @classmethod
    async def createAndRun_async(cls, *args, **kwargs):
        self = cls(*args, **kwargs)
        while True:
            await asyncio.sleep(1)

    @classmethod
    def createAndRun(cls, *args, **kwargs):
        asyncio.run(cls.createAndRun_async(*args, **kwargs))


@attrs.define()
class ThreadedLSLStreamResolver(LSLStreamResolver):
    _pollPeriod: float = 1.
    _subPort: tp.Optional[int] = None
    _reqPort: tp.Optional[int] = None

    _thread: threading.Thread = attrs.field(init=False)
    _connector: ZMQConnectorClient = attrs.field(init=False)

    def __attrs_post_init__(self):
        LSLStreamResolver.__attrs_post_init__(self)

        if self._subPort is None:
            self._subPort = getNewPort()

        if self._reqPort is None:
            self._reqPort = getNewPort()

        self._connector = ZMQConnectorClient(
            pubSubPort=self._subPort,
            reqRepPort=self._reqPort,
            onMessagePublished=self._onMessagePublished,
            allowAsyncCalls=True
        )

        self._thread = threading.Thread(target=_ThreadedLSLStreamResolver_Daemon.createAndRun, kwargs=dict(
            pollPeriod=self._pollPeriod,
            pubPort=self._subPort,
            repPort=self._reqPort,
            resolveStreamsTimeout=self._resolveStreamsTimeout
        ), daemon=True)
        self._thread.start()
        asyncio.create_task(asyncTryAndLogExceptionOnError(self._startPolling))

    async def _startPolling(self):
        await asyncio.sleep(self._pollPeriod)  # extra delay here to give time for subscribe to take effect
                                                # TODO: try shortening this delay
        await self._connector.call_async('startAsyncPolling')

    def updateAvailableStreams(self):
        raise NotImplementedError()  # this happens in remote thread, should not be called here

    def _onMessagePublished(self, msg: tp.List[bytes]):
        logger.debug('Received message of type %s' % (msg[0].decode('utf-8'),))
        if msg[0] == b'streamDetected':
            d = json.loads(msg[1].decode('utf-8'))
            streamKey = d['key']
            timeout = 0.01
            infos = []
            while timeout < 10.:  # TODO: don't hardcode this limit
                infos = lsl.resolve_bypred("name='%s' and hostname='%s' and type='%s' and source_id='%s'" %
                                           (d['name'], d['hostname'], d['type'], d['sourceID']),
                                           minimum=1, timeout=timeout)
                if len(infos) == 0:
                    # timed out
                    timeout *= 2  # try with longer timeout
                else:
                    break

            if len(infos) == 0:
                # timed out even after trying long timeout
                logger.error('Remote reported streamDetected but could not resolve, dropping.')
                return

            if len(infos) > 1:
                # multiple streams resolved when we only expected 1
                logger.error('Expected to resolve one stream, but found multiple matches.')
                raise NotImplementedError()

            streamInfo = infos[0]
            self._onStreamDetected(streamKey=streamKey, streamInfo=streamInfo)

        elif msg[0] == b'streamLost':
            streamKey = msg[1].decode('utf-8')
            self._onStreamLost(streamKey=streamKey)

        else:
            logger.error('Unexpected message')
            raise NotImplementedError()

    def markStreamAsLost(self, streamKey: str):
        self._connector.call('markStreamAsLost', streamKey=streamKey)





from __future__ import annotations
import typing as tp
import pyqtgraph.dockarea as pgd
from pyqtgraph.dockarea.Dock import DockLabel as pgdDockLabel, VerticalLabel as pgdVerticalLabel
import pyqtgraph.dockarea.DockDrop as pgdd
import pyqtgraph.dockarea.Container as pgdc
from qtpy import QtWidgets, QtGui, QtCore
import weakref

from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.GUI.StyleSheets import setStyleSheetForInstanceOnly



borderWidth = '2px'


class LabelScrollArea(QtWidgets.QScrollArea):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)

    def sizeHint(self) -> QtCore.QSize:
        """
        Modify size hint behavior to prevent scrollbar showing in some situations where the widget could be given more room to grow
        """
        sz = self.widget().sizeHint()
        sz.setWidth(sz.width()+10)
        return sz


class AltTContainer(QtWidgets.QTabWidget, pgdc.TContainer):
    def __init__(self, area):
        QtWidgets.QTabWidget.__init__(self)
        pgdc.Container.__init__(self, area)

        if True:

            r = '5px'
            if self.parent() is None:
                palette = self.palette()
            else:
                palette = self.parent().palette()

            fg = '#444444'
            bg = '#dddddd'
            borderColor = '#bbbbbb'
            border = borderColor
            borderBottom = borderColor
            thisBorderWidth = '0px'

            setStyleSheetForInstanceOnly(self.tabBar(), f"""
                background-color : {bg};
                color : {fg};
                border-top-right-radius: {r};
                border-top-left-radius: {r};
                border-bottom-right-radius: 0px;
                border-bottom-left-radius: 0px;
                border-width: {thisBorderWidth};
                border-bottom: {thisBorderWidth} solid {borderBottom};
                border-top: {thisBorderWidth} solid {border};
                border-left: {thisBorderWidth} solid {border};
                border-right: {thisBorderWidth} solid {border};
                padding-left: 1px;
                padding-right: 1px;
                """, selectorSuffix='::tab')

            fg = palette.color(QtGui.QPalette.Active, QtGui.QPalette.Text).name()
            bg = '#bbbbbb'
            borderColor = '#bbbbbb'
            border = borderColor
            borderBottom = bg
            thisBorderWidth = borderWidth

            setStyleSheetForInstanceOnly(self.tabBar(), f"""
                            background-color : {bg};
                            color : {fg};
                            border-width: {thisBorderWidth};
                            border-bottom: {thisBorderWidth} solid {borderBottom};
                            border-top: {thisBorderWidth} solid {border};
                            border-left: {thisBorderWidth} solid {border};
                            border-right: {thisBorderWidth} solid {border};
                            """,
                                         selectorSuffix='::tab::selected')
        else:
            self.setStyleSheet(f"""
                QTabBar::tab:selected {{ 
                  background: #bbbbbb; 
                }}
                """)



    def _insertItem(self, item, index):
        if not isinstance(item, Dock):
            raise Exception("Tab containers may hold only docks, not other containers.")

        self.insertTab(index,
                       item,
                       item.label.icon.pixmap(),
                       item.label.text)

    def raiseDock(self, dock):
        index = self.indexOf(dock)
        self.setCurrentIndex(index)


class StackedWidget(pgdc.StackedWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.currentChanged.connect(lambda *args: self._onCurrentChanged(*args))

    def sizeHint(self) -> QtCore.QSize:
        """
        Modify size hint to return current widget size hint instead of maximum size
        """
        if self.currentWidget() is not None:
            return self.currentWidget().sizeHint()
        else:
            return super().sizeHint()

    def minimumSizeHint(self) -> QtCore.QSize:
        """
        Modify size hint to return current widget size hint instead of maximum size
        """
        if self.currentWidget() is not None:
            return self.currentWidget().minimumSizeHint()
        else:
            return super().minimumSizeHint()

    def _onCurrentChanged(self, index: int):
        if self.currentWidget() is not None:
            self.setSizePolicy(self.currentWidget().sizePolicy())
            self.adjustSize()

    def childEvent(self, ev):
        super().childEvent(ev)
        self.container.childEvent_(ev)


class TContainer(pgdc.TContainer):
    def __init__(self, area):
        QtWidgets.QWidget.__init__(self)
        pgdc.Container.__init__(self, area)
        self.layout = QtWidgets.QGridLayout()
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        self.hTabLayout = QtWidgets.QHBoxLayout()
        self.hTabBox = QtWidgets.QWidget()
        self.hTabBox.setLayout(self.hTabLayout)
        self.hTabLayout.setSpacing(2)
        self.hTabLayout.setContentsMargins(0, 0, 0, 0)

        if False:
            self.hTabBoxScroll = LabelScrollArea()
            setStyleSheetForInstanceOnly(self.hTabBoxScroll, 'background: transparent;')
            self.hTabBox.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
            self.hTabBoxScroll.setWidgetResizable(True)
            self.hTabBoxScroll.setWidget(self.hTabBox)
            self.hTabBoxScroll.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Maximum)
            self.layout.addWidget(self.hTabBoxScroll, 0, 1)
        else:
            self.layout.addWidget(self.hTabBox, 0, 1)

        self.stack = StackedWidget(container=self)
        self.layout.addWidget(self.stack, 1, 1)
        self.stack.currentChanged.connect(self._onStackCurrentChanged)

        self.setLayout(self.layout)
        for n in ['count', 'widget', 'indexOf']:
            setattr(self, n, getattr(self.stack, n))

    def _insertItem(self, item, index):
        prevIndex = self.stack.currentIndex()
        if index <= prevIndex or prevIndex == -1:
            prevIndex += 1
        if not isinstance(item, Dock):
            raise Exception("Tab containers may hold only docks, not other containers.")
        self.stack.insertWidget(index, item)
        self.hTabLayout.insertWidget(index, item.label)
        #QtCore.QObject.connect(item.label, QtCore.SIGNAL('clicked'), self.tabClicked)
        item.label.sigClicked.connect(self.tabClicked)
        if True:
            # don't show new item by default
            item.label.setDim(True)
            self.raiseDock(self.stack.widget(prevIndex))
        else:
            self.tabClicked(item.label)

    def _onStackCurrentChanged(self, index: int):
        if self.stack.currentWidget() is not None:
            self.setSizePolicy(self.stack.currentWidget().sizePolicy())
            self.adjustSize()

    def restoreState(self, state):
        super().restoreState(state)
        self.tabClicked(self.stack.widget(state['index']).label)


class HContainer(pgdc.HContainer):
    def __init__(self, area):
        super().__init__(area)

        if False:
            self.setStyleSheet(f"""
                QSplitter::handle {{
                    background-color: '#ff0000';
                }}
                QSplitter::handle:horizontal {{
                    width: 4px;
                }}
                QSplitter::handle:vertical {{
                    height: 4px;
                }}
            """)

    def _insertItem(self, item, index):
        super()._insertItem(item, index)
        self.setCollapsible(index, False)  # don't allow collapsing


class VContainer(pgdc.VContainer):
    def __init__(self, area):
        super().__init__(area)

        if False:
            self.setStyleSheet(f"""
                QSplitter::handle {{
                    background-color: '#ff0000';
                }}
                QSplitter::handle:horizontal {{
                    width: 4px;
                }}
                QSplitter::handle:vertical {{
                    height: 4px;
                }}
            """)


    def _insertItem(self, item, index):
        super()._insertItem(item, index)
        self.setCollapsible(index, False)  # don't allow collapsing


class DockLabel(QtWidgets.QFrame):
    _icon: QtWidgets.QLabel | None = None
    _iconSrc: QtGui.QIcon | None = None
    _iconFn: tp.Callable[..., QtGui.QIcon] | None = None
    _label: QtWidgets.QLabel
    _closeButton: QtWidgets.QToolButton | None = None
    _dock: Dock
    _dim: bool = False
    _iconSize: tuple[int, int] | None = None

    _styleUpdateInProgress: bool = False

    sigClicked = QtCore.Signal(object, object)
    sigCloseClicked = QtCore.Signal()

    def __init__(self, text: str, dock: Dock, showCloseButton: bool,
                 icon: QtGui.QIcon | None = None,
                 iconFn: tp.Callable[..., QtGui.QIcon] | None = None,
                 iconSize: tuple[int, int] | None = None,
                 **kwargs):
        QtWidgets.QFrame.__init__(self)
        self._dock = dock
        self._iconSize = iconSize
        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(1, 1, 1, 1)
        self.layout().setSpacing(1)

        if icon is not None or iconFn is not None:
            self._iconSrc = icon
            self._iconFn = iconFn
            self._icon = QtWidgets.QLabel(self)
            self.layout().addWidget(self._icon)
            self._icon.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)

        self._label = QtWidgets.QLabel(text, self)
        self._label.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        self.layout().addWidget(self._label)

        if showCloseButton:
            self._closeButton = QtWidgets.QToolButton(self)
            self._closeButton.setIcon(QtWidgets.QApplication.style().standardIcon(QtWidgets.QStyle.StandardPixmap.SP_TitleBarCloseButton))
            self._closeButton.clicked.connect(self.sigCloseClicked)
            self.layout().addWidget(self._closeButton)

        self.updateStyle()

        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)

    @property
    def dim(self):
        return self._dim

    @property
    def icon(self):
        return self._icon

    @property
    def text(self):
        return self._label.text()

    def setDim(self, d):
        if self.dim != d:
            self._dim = d
            self.updateStyle()

    @property
    def dock(self):
        return self._dock

    def setOrientation(self, o: str):
        if o == 'auto':
            return
        elif o == 'vertical':
            raise NotImplementedError
        elif o == 'horizontal':
            return
        else:
            raise NotImplementedError

    def event(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.Type.Polish:
            self.updateStyle()
        elif event.type() == QtCore.QEvent.Type.PaletteChange:
            if not self._styleUpdateInProgress:
                QtCore.QTimer.singleShot(0, self.updateStyle)
        elif event.type() == QtCore.QEvent.Type.ApplicationPaletteChange:
            self.updateStyle()
        return super().event(event)

    def updateStyle(self):
        self._styleUpdateInProgress = True

        if self._icon is not None:
            if self._iconFn is not None:
                icon = self._iconFn()
            else:
                icon = self._iconSrc

            iconSize = self._iconSize
            if iconSize is None:
                # calculate iconSize from font size
                iconSize = tuple(1.25 * self._label.fontMetrics().height() for i in range(2))
            self._icon.setPixmap(icon.pixmap(*iconSize))

        r = '5px'
        if self.parent() is None:
            palette = self.palette()
        else:
            palette = self.parent().palette()
        if self.dim:
            self._label.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
            self._label.setMinimumWidth(1)
            if palette.color(QtGui.QPalette.Base).value() > 128:
                # light theme
                fg = '#333333'
                bg = '#dddddd'
                hbg = '#cccccc'
                borderColor = '#bbbbbb'
            else:
                # dark theme
                fg = '#cccccc'
                bg = '#444444'
                hbg = '#555555'
                borderColor = '#666666'

            border = borderColor
            borderBottom = borderColor
            thisBorderWidth = '0px'
        else:
            self._label.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Maximum)
            self._label.setMinimumWidth(0)

            if palette.color(QtGui.QPalette.Base).value() > 128:
                # light theme
                fg = palette.color(QtGui.QPalette.Active, QtGui.QPalette.Text).name()
                bg = '#bbbbbb'
                borderColor = '#bbbbbb'
            else:
                # dark theme
                fg = '#eeeeee'
                bg = '#666666'
                borderColor = '#666666'

            hbg = bg
            border = borderColor
            borderBottom = bg
            thisBorderWidth = borderWidth

        self.hStyle = f"""
            background-color : {bg};
            color : {fg};
            border-top-right-radius: {r};
            border-top-left-radius: {r};
            border-bottom-right-radius: 0px;
            border-bottom-left-radius: 0px;
            border-width: {thisBorderWidth};
            border-bottom: {thisBorderWidth} solid {borderBottom};
            border-top: {thisBorderWidth} solid {border};
            border-left: {thisBorderWidth} solid {border};
            border-right: {thisBorderWidth} solid {border};
            padding-left: 1px;
            padding-right: 1px;
        """
        if False:
            # (for some reason, this causes crash when mouse hovers over the label, so disabled for now)
            setStyleSheetForInstanceOnly(self, f"""
                background-color: {hbg};
                """, selectorSuffix=':hover')
        setStyleSheetForInstanceOnly(self, self.hStyle)

        setStyleSheetForInstanceOnly(self._label, f""" 
            font-weight: {400 if self.dim else 500};
            color: {fg};
            """)
        self._styleUpdateInProgress = False

    def mousePressEvent(self, ev):
        lpos = ev.position() if hasattr(ev, 'position') else ev.localPos()
        self.pressPos = lpos
        self.mouseMoved = False
        ev.accept()

    def mouseMoveEvent(self, ev):
        if not self.mouseMoved:
            lpos = ev.position() if hasattr(ev, 'position') else ev.localPos()
            self.mouseMoved = (lpos - self.pressPos).manhattanLength() > QtWidgets.QApplication.startDragDistance()

        if self.mouseMoved and ev.buttons() == QtCore.Qt.MouseButton.LeftButton:
            self.dock.startDrag()
        ev.accept()

    def mouseReleaseEvent(self, ev):
        ev.accept()
        if not self.mouseMoved:
            self.sigClicked.emit(self, ev)

    def mouseDoubleClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            self.dock.float()


class DockDrop(pgdd.DockDrop):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def dragEnterEvent(self, ev):
        src = ev.source()
        if isinstance(src, Dock):
            if isinstance(self.dndWidget, DockArea):
                existingAffinities: set = set()
                for dock in self.dndWidget.docks.values():
                    existingAffinities |= set(dock.affinities if dock.affinities is not None else [None])
            elif isinstance(self.dndWidget, Dock):
                existingAffinities: set = set(self.dndWidget.affinities if self.dndWidget.affinities is not None else [None])
            else:
                raise NotImplementedError

            if len(existingAffinities) == 0:
                doesMatch = True

            else:
                if src.affinities is None:
                    doesMatch = None in existingAffinities
                else:
                    doesMatch = len(existingAffinities & set(src.affinities)) > 0

            if doesMatch:
                ev.accept()
            else:
                ev.ignore()
        else:
            ev.ignore()


class Dock(pgd.Dock):
    _affinities: list[str] | None = None

    _styleUpdateInProgress: bool = False

    sigFocused: Signal
    """
    This signal must be emitted by slot connected to root focusObjectChanged to function.
    """
    sigShown: Signal[()]
    sigHidden: Signal[()]
    sigResized: Signal[()]

    # noinspection PyMissingConstructor
    def __init__(self, name, title: str = None, area=None, size=(10, 10), widget=None, hideTitle=False, autoOrientation=False, closable=False, affinities: list[str] | None = None,
                 icon: QtGui.QIcon | None = None,
                 iconFn: tp.Callable[..., QtGui.QIcon] | None = None):
        # completely override parent class init to specify different DockDrop class

        QtWidgets.QWidget.__init__(self)
        self.dockdrop = DockDrop(self)
        self._container = None
        self._name = name
        self.area = area
        if title is None:
            title = self._name
        self.label = (
            DockLabel(title, self, closable, icon=icon, iconFn=iconFn))
        if closable:
            self.label.sigCloseClicked.connect(self.close)
        self.labelHidden = False
        self.moveLabel = True  ## If false, the dock is no longer allowed to move the label.
        self.autoOrient = autoOrientation
        self.orientation = 'horizontal'
        # self.label.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter)
        self.topLayout = QtWidgets.QGridLayout()
        self.topLayout.setContentsMargins(0, 0, 0, 0)
        self.topLayout.setSpacing(0)
        self.setLayout(self.topLayout)
        self.topLayout.addWidget(self.label, 0, 1)
        self.widgetArea = QtWidgets.QWidget()
        self.topLayout.addWidget(self.widgetArea, 1, 1)
        self.layout = QtWidgets.QGridLayout()
        self.layout.setContentsMargins(2, 2, 2, 2)
        self.layout.setSpacing(0)
        self.widgetArea.setLayout(self.layout)
        #self.widgetArea.setSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        self.widgets = []
        self.currentRow = 0
        # self.titlePos = 'top'
        self.dockdrop.raiseOverlay()

        self.setStretch(*size)

        if widget is not None:
            self.addWidget(widget)

        if hideTitle:
            self.hideTitleBar()

        if affinities is not None:
            self._affinities = affinities

        self.sigFocused = Signal()
        self.sigShown = Signal()
        self.sigHidden = Signal()
        self.sigResized = Signal()

        self.updateStyle()

    def event(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.Type.Polish:
            self.updateStyle()
        elif event.type() == QtCore.QEvent.Type.PaletteChange:
            if not self._styleUpdateInProgress:
                QtCore.QTimer.singleShot(0, self.updateStyle)
        elif event.type() == QtCore.QEvent.Type.ApplicationPaletteChange:
            self.updateStyle()
        return super().event(event)

    def updateStyle(self):
        self._styleUpdateInProgress = True

        if self.parent() is None:
            palette = self.palette()
        else:
            palette = self.parent().palette()
        if palette.color(QtGui.QPalette.Base).value() > 128:
            # light theme
            borderColor = '#bbbbbb'
        else:
            # dark theme
            borderColor = '#666666'

        self.hStyle = f"""
        Dock > QWidget {{
            border: {borderWidth} solid {borderColor};
            border-radius: 5px;
            border-top-left-radius: 0px;
            border-top-right-radius: 0px;
            border-top-width: {borderWidth};
        }}"""
        self.vStyle = f"""
        Dock > QWidget {{
            border: {borderWidth} solid {borderColor};
            border-radius: 5px;
            border-top-left-radius: 0px;
            border-bottom-left-radius: 0px;
            border-left-width: 0px;
        }}"""
        self.nStyle = """
        Dock > QWidget {
            border: 2px solid #000;
            border-radius: 5px;
        }"""
        self.dragStyle = """
        Dock > QWidget {
            border: 4px solid #00F;
            border-radius: 5px;
        }"""
        self.setAutoFillBackground(False)
        self.widgetArea.setStyleSheet(self.hStyle)

        super().updateStyle()

        self._styleUpdateInProgress = False

    def setEnabled(self, doEnable: bool):
        self.label.setEnabled(doEnable)
        super().setEnabled(doEnable)

    def implements(self, name=None):
        if name is None:
            return ['dock', 'dockWithAffinities']
        else:
            return name in ('dock', 'dockWithAffinities')

    @property
    def affinities(self):
        return self._affinities

    def dragEnterEvent(self, *args):
        self.dockdrop.dragEnterEvent(*args)

    def dragMoveEvent(self, *args):
        self.dockdrop.dragMoveEvent(*args)

    def dragLeaveEvent(self, *args):
        self.dockdrop.dragLeaveEvent(*args)

    def dropEvent(self, *args):
        self.dockdrop.dropEvent(*args)

    def raiseDock(self):
        if self.container() is not None and self.container().type() == 'tab':
            super().raiseDock()

    def hideEvent(self, *args):
        super().hideEvent(*args)
        self.sigHidden.emit()

    def showEvent(self, *args):
        super().showEvent(*args)
        self.sigShown.emit()

    def resizeEvent(self, *args):
        super().resizeEvent(*args)
        self.sigResized.emit()


class DockArea(pgd.DockArea):
    # noinspection PyMissingConstructor

    _affinities: list[str] | None = None

    def __init__(self, parent=None, temporary=False, home=None, affinities: list[str] | None = None):
        # completely override parent class init to specify different DockDrop class

        pgdc.Container.__init__(self, self)
        QtWidgets.QWidget.__init__(self, parent=parent)
        self.dockdrop = DockDrop(self)
        self.dockdrop.removeAllowedArea('center')
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.setLayout(self.layout)
        self.docks = weakref.WeakValueDictionary()
        self.topContainer = None
        self.dockdrop.raiseOverlay()
        self.temporary = temporary
        self.tempAreas = []
        self.home = home
        self._affinities = affinities

    @property
    def affinities(self):
        return self._affinities

    def addDock(self, dock: Dock, position='bottom',
                relativeTo: Dock | None = None):
        if self._affinities is not None:
            assert dock.affinities is not None
            assert any(x in self._affinities for x in dock.affinities)

        super().addDock(dock, position=position, relativeTo=relativeTo)
        if dock.affinities is not None:
            for affinity in dock.affinities:
                if affinity in self.docks:
                    self.moveDock(dock, 'above', self.docks[affinity])
                    break

    def dragEnterEvent(self, *args):
        self.dockdrop.dragEnterEvent(*args)

    def dragMoveEvent(self, *args):
        self.dockdrop.dragMoveEvent(*args)

    def dragLeaveEvent(self, *args):
        self.dockdrop.dragLeaveEvent(*args)

    def dropEvent(self, *args):
        self.dockdrop.dropEvent(*args)

    def makeContainer(self, typ):
        if typ == 'vertical':
            new = VContainer(self)
        elif typ == 'horizontal':
            new = HContainer(self)
        elif typ == 'tab':
            if True:
                new = TContainer(self)
            else:
                new = AltTContainer(self)
        else:
            raise ValueError("typ must be one of 'vertical', 'horizontal', or 'tab'")
        return new

    def buildFromState(self, state, docks, root, depth=0, missing='error'):
        # override parent to disable apoptosing that can cause issues in some situations

        typ, contents, state = state
        pfx = "  " * depth
        if typ == 'dock':
            try:
                obj = docks[contents]
                del docks[contents]
            except KeyError:
                if missing == 'error':
                    raise Exception('Cannot restore dock state; no dock with name "%s"' % contents)
                elif missing == 'create':
                    obj = Dock(name=contents)
                elif missing == 'ignore':
                    return
                else:
                    raise ValueError('"missing" argument must be one of "error", "create", or "ignore".')

        else:
            obj = self.makeContainer(typ)

        root.insert(obj, 'after')
        # print pfx+"Add:", obj, " -> ", root

        if typ != 'dock':
            for o in contents:
                self.buildFromState(o, docks, obj, depth + 1, missing=missing)
            if False:
                # remove this container if possible. (there are valid situations when a restore will
                # generate empty containers, such as when using missing='ignore')
                obj.apoptose(propagate=False)
            obj.restoreState(state)  ## this has to be done later?

    def restoreState(self, state, missing='error', extraPosition='bottom', extraNeighbor=None):
        """
        Similar to superclass, but allows specifying neighbor for extra relative position
        """
        ## 1) make dict of all docks and list of existing containers
        containers, docks = self.findAll()
        oldTemps = self.tempAreas[:]
        # print "found docks:", docks

        ## 2) create container structure, move docks into new containers
        if state['main'] is not None:
            self.buildFromState(state['main'], docks, self, missing=missing)

        ## 3) create floating areas, populate
        for s in state['float']:
            a = self.addTempArea()
            a.buildFromState(s[0]['main'], docks, a, missing=missing)
            a.win.setGeometry(*s[1])
            a.apoptose()  # ask temp area to close itself if it is empty

        ## 4) Add any remaining docks to a float
        for d in docks.values():
            if extraPosition == 'float':
                assert extraNeighbor is None
                a = self.addTempArea()
                a.addDock(d, 'below')
            else:
                self.moveDock(d, extraPosition, extraNeighbor)

        # print "\nKill old containers:"
        ## 5) kill old containers
        for c in containers:
            c.close()
        for a in oldTemps:
            a.apoptose()





from __future__ import annotations

import asyncio

import attrs
import contextlib
import darkdetect
import logging
import os
import pathlib
import platformdirs
import pyvista as pv
import qtawesome as qta
from qtpy import QtGui, QtCore, QtWidgets
import typing as tp

from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.GUI.QAppWithAsyncioLoop import RunnableAsApp
from NaviNIBS.util.GUI.Dock import DockArea
from NaviNIBS.util.GUI.ErrorDialog import asyncTryAndRaiseDialogOnError
from NaviNIBS.util.logging import getLogFilepath, createLogFileHandler
from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.Navigator.Model.DockWidgetLayouts import DockWidgetLayout
from NaviNIBS.Navigator.GUI.ViewPanels import MainViewPanel
from NaviNIBS.Navigator.GUI.ViewPanels.MainViewPanelWithDockWidgets import MainViewPanelWithDockWidgets
from NaviNIBS.Navigator.GUI.ViewPanels.ManageSessionPanel import ManageSessionPanel
from NaviNIBS.Navigator.GUI.ViewPanels.MRIPanel import MRIPanel
from NaviNIBS.Navigator.GUI.ViewPanels.HeadModelPanel import HeadModelPanel
from NaviNIBS.Navigator.GUI.ViewPanels.FiducialsPanel import FiducialsPanel
from NaviNIBS.Navigator.GUI.ViewPanels.TargetsPanel import TargetsPanel
from NaviNIBS.Navigator.GUI.ViewPanels.ToolsPanel import ToolsPanel
from NaviNIBS.Navigator.GUI.ViewPanels.TriggerSettingsPanel import TriggerSettingsPanel
from NaviNIBS.Navigator.GUI.ViewPanels.CameraPanel import CameraPanel
from NaviNIBS.Navigator.GUI.ViewPanels.SubjectRegistrationPanel import SubjectRegistrationPanel
from NaviNIBS.Navigator.GUI.ViewPanels.NavigatePanel import NavigatePanel
from NaviNIBS.Navigator.GUI.ViewPanels.DigitizeLocationsPanel import DigitizeLocationsPanel

logger = logging.getLogger(__name__)


_VP = tp.TypeVar('_VP', bound=MainViewPanel)


@attrs.define
class NavigatorGUI(RunnableAsApp):
    _appName: str = 'NaviNIBS Navigator GUI'

    _sesFilepath: tp.Optional[str] = None  # only used to load session on startup
    _inProgressBaseDir: tp.Optional[str] = None

    _session: tp.Optional[Session] = None

    _rootDockArea: DockArea = attrs.field(init=False)

    _mainViewPanels: tp.Dict[str, MainViewPanel] = attrs.field(init=False, factory=dict)

    _logFileHandler: tp.Optional[logging.FileHandler] = attrs.field(init=False, default=None)
    _theme: str = 'light'
    """
    Default theme. Can be overriden by field in Session.miscSettings
    """

    _restoringLayoutLock: asyncio.Lock = attrs.field(init=False, factory=asyncio.Lock)

    def __attrs_post_init__(self):
        logger.info('Initializing {}'.format(self.__class__.__name__))

        thisDir = pathlib.Path(__file__).parent
        iconPath = os.path.join(thisDir, 'resources', 'NaviNIBSIcon.ico')
        assert os.path.exists(iconPath)
        self._appIconPath = iconPath

        super().__attrs_post_init__()

        self._app.setAttribute(QtCore.Qt.ApplicationAttribute.AA_UseStyleSheetPropagationInWidgetStyles, True)

        if self._inProgressBaseDir is None:
            self._inProgressBaseDir = os.path.join(platformdirs.user_data_dir(appname='NaviNIBS', appauthor=False), 'InProgressSessions')

        self._rootDockArea = DockArea(affinities=['MainViewPanel'])
        self._rootDockArea.setContentsMargins(2, 2, 2, 2)
        self._win.setCentralWidget(self._rootDockArea)

        self._refreshGUIAppearance()

        panel = self._addViewPanel(ManageSessionPanel(key='Manage session',
                                                      session=self._session,
                                                      inProgressBaseDir=self._inProgressBaseDir,
                                                      navigatorGUI=self),)
        panel.sigAboutToFinishLoadingSession.connect(self._onSessionAboutToFinishLoading)
        panel.sigLoadedSession.connect(self._onSessionLoaded)
        panel.sigClosedSession.connect(self._onSessionClosed)
        panel.addSaveButtonToDockTabStrip(dockArea=self._rootDockArea)

        self._addViewPanel(MRIPanel(session=self._session))

        self._addViewPanel(HeadModelPanel(session=self._session))

        self._addViewPanel(FiducialsPanel(session=self._session))

        #self._addViewPanel(CoordinateSystemsPanel(session=self._session))
        # TODO: set up transforms widget

        self._addViewPanel(TargetsPanel(session=self._session))

        #self._toolbarWdgt.addSeparator()  # separate pre-session planning/setup panels from within-session panels

        self._addViewPanel(ToolsPanel(session=self._session))

        self._addViewPanel(TriggerSettingsPanel(session=self._session))

        self._addViewPanel(CameraPanel(session=self._session))

        self._addViewPanel(SubjectRegistrationPanel(session=self._session))

        #self._toolbarWdgt.addSeparator()  # separate pre-session planning/setup panels from within-session panels

        self._addViewPanel(NavigatePanel(session=self._session))

        self._addViewPanel(DigitizeLocationsPanel(session=self._session))

        # set initial view widget visibility
        # TODO: default to MRI if new session, otherwise default to something else...
        self._updateEnabledPanels()
        self._activateView('Manage session')
        #self._activateView('Navigate')  # TODO: debug, delete

        if self._sesFilepath is not None:
            asyncio.create_task(asyncTryAndRaiseDialogOnError(self._loadAfterSetup, filepath=self._sesFilepath))

        if self._doRunAsApp:
            logger.debug('Showing window')
            self._win.resize(QtCore.QSize(1600, 1100))  # TODO: restore previous size if available
            self._win.show()

    def _addViewPanel(self, panel: _VP) -> _VP:
        logger.info(f'Adding view panel {panel.key}')
        if len(self._mainViewPanels) > 0:
            relTo = self._mainViewPanels[list(self._mainViewPanels.keys())[-1]].dockWdgt
        else:
            relTo = None

        self._rootDockArea.addDock(panel.dockWdgt, position='below',
                                         relativeTo=relTo)
        self._mainViewPanels[panel.key] = panel
        if isinstance(panel, MainViewPanelWithDockWidgets):
            panel.sigAboutToRestoreLayout.connect(self._onAboutToRestorePanelLayout)
            panel.sigRestoredLayout.connect(self._onRestoredPanelLayout)
        return panel

    def _saveRootLayout(self):
        # save root layout

        key = 'NavigatorGUI'
        layout = self.session.dockWidgetLayouts.get(key, None)
        if layout is None:
            layout = DockWidgetLayout(
                key=key,
                affinities=['MainViewPanel'],
            )
            self.session.dockWidgetLayouts.addItem(layout)

        sz = self._win.size()
        layout.winSize = (sz.width(), sz.height())
        layout.state = self._rootDockArea.saveState()

    def saveLayout(self):
        self._saveRootLayout()

        # save each initialized panel's layout
        for mainViewPanel in self.mainViewPanels.values():
            if isinstance(mainViewPanel, MainViewPanelWithDockWidgets) and mainViewPanel.hasInitialized:
                mainViewPanel.saveLayout()

    async def _restoreRootLayout(self, needsLock: bool = True):
        # restore root layout

        await asyncio.sleep(0.01)
        layout = self.session.dockWidgetLayouts.get('NavigatorGUI', None)

        if layout is not None and layout.winSize is not None:
            self._win.resize(QtCore.QSize(*layout.winSize))

        if layout is not None and layout.state is not None:

            async with self._restoringLayoutLock if needsLock else contextlib.nullcontext():

                with contextlib.ExitStack() as stack:
                    blocks = [stack.enter_context(panel.sigPanelShown.blocked()) for panel in self.mainViewPanels.values()]

                    winSize = self._win.size()
                    logger.debug(f'About to restore root layout when winSize={winSize}')

                    try:
                        self._rootDockArea.restoreState(layout.state,
                                                        extraPosition='below',
                                                        extraNeighbor=self.manageSessionPanel.dockWdgt)
                    except ValueError:
                        # sometimes can get errors during restore if other parts of layout have changed
                        logger.error('Unable to restore root dock area state')

                for panel in self.mainViewPanels.values():
                    # catch up with signals that may have been blocked above
                    await asyncio.sleep(0.01)
                    panel.wdgt.updateGeometry()
                    # if panel.dockWdgt.isCurrentTab():
                    #     panel.sigPanelShown.emit()
                    # else:
                    #     panel.sigPanelHidden.emit()

                await asyncio.sleep(0.01)
                winSize = self._win.size()
                logger.debug(f'Restored root layout when winSize={winSize}')
                # jiggle window size to force layout to update
                # TODO: find way to force layout update without jiggle that is visible to user
                self._win.resize(winSize - QtCore.QSize(10, 10))  # TODO: restore previous size if available
                await asyncio.sleep(0.01)
                self._win.resize(winSize)
                await asyncio.sleep(0.01)
                logger.debug(f'After restored root layout, winSize={winSize}')

    def _onAboutToRestorePanelLayout(self):
        if self._restoringLayoutLock.locked():
            return  # ignore if in the middle of a whole-app restore
        self._saveRootLayout()

    def _onRestoredPanelLayout(self):
        if self._restoringLayoutLock.locked():
            return  # ignore if in the middle of a whole-app restore
        asyncio.create_task(asyncTryAndLogExceptionOnError(self._restoreRootLayout))

    async def restoreLayoutIfAvailable(self):
        async with self._restoringLayoutLock:
            await self._restoreRootLayout(needsLock=False)

            # restore each initialized panel's layout
            didRestoreAPanel = False
            for mainViewPanel in self.mainViewPanels.values():
                if isinstance(mainViewPanel, MainViewPanelWithDockWidgets) and mainViewPanel.hasInitialized:
                    if hasattr(mainViewPanel, 'finishedAsyncInitialization'):
                        if True:
                            if not mainViewPanel.finishedAsyncInitialization.is_set():
                                continue  # assume it will restore itself when initialized
                        else:
                            await mainViewPanel.finishedAsyncInitialization.wait()

                        continue  # TODO: debug, delete
                    if mainViewPanel.restoreLayoutIfAvailable():
                        didRestoreAPanel = True

            if didRestoreAPanel and False:
                # restore root layout again
                await self._restoreRootLayout(needsLock=False)

    def _onAppAboutToQuit(self):
        super()._onAppAboutToQuit()

        # close each non-visible panel first to prevent them from initializing right before closing
        for panelKey, panel in self._mainViewPanels.items():
            if not panel.isVisible:
                panel.close()

    def _onSessionAboutToFinishLoading(self, session: Session):
        if self._logFileHandler is not None:
            # remove previous session log file handler
            logging.getLogger('').removeHandler(self._logFileHandler)
            self._logFileHandler = None
        self._logFileHandler = createLogFileHandler(getLogFilepath(session))
        logging.getLogger('').addHandler(self._logFileHandler)

    def _onSessionLoaded(self, session: Session):
        assert session is not None
        logger.info('Loaded session {}'.format(session.filepath))

        self._onAddonsAboutToChange(session.addons.keys())

        self._session = session

        for pane in self._mainViewPanels.values():
            pane.session = session

        self._onAddonsChanged(session.addons.keys(), triggeredBySessionLoad=True)

        self._refreshGUIAppearance()

        asyncio.create_task(asyncTryAndLogExceptionOnError(self.restoreLayoutIfAvailable))

        self._updateEnabledPanels()
        session.MRI.sigFilepathChanged.connect(self._updateEnabledPanels)
        session.headModel.sigFilepathChanged.connect(self._updateEnabledPanels)
        session.addons.sigItemsAboutToChange.connect(lambda *args: self._onAddonsAboutToChange(*args))
        session.addons.sigItemsChanged.connect(lambda *args: self._onAddonsChanged(*args, triggeredBySessionLoad=False))

        self.session.subjectRegistration.fiducials.sigItemsChanged.connect(lambda *args: self._updateEnabledPanels())
        self.session.tools.sigItemsChanged.connect(lambda *args: self._updateEnabledPanels())

        self.session.miscSettings.sigAttribsChanged.connect(self._onSessionMiscSettingsChanged)

    def _refreshGUIAppearance(self):
        if self._session is None:
            return

        logger.debug('Refreshing GUI appearance')

        theme = self._session.miscSettings.theme.lower()
        if theme == 'auto':
            theme = 'dark' if darkdetect.isDark() else 'light'
        self.theme = theme  # triggers RunnableAsApp theme setter
        match theme:
            case 'light':
                pv.set_plot_theme('default')
            case 'dark':
                pv.set_plot_theme('dark')
            case _:
                raise ValueError(f'Unknown theme: {theme}')

        pv.global_theme.background = self._win.palette().color(QtGui.QPalette.Base).name()
        pv.global_theme.font.color = self._win.palette().color(QtGui.QPalette.Text).name()

        mainFontSize = self._session.miscSettings.mainFontSize
        if mainFontSize is None:
            styleContent = ''
        else:
            styleContent = f'font-size: {mainFontSize}pt;'

        self._app.setStyleSheet(styleContent)
        for win in self._app.topLevelWidgets():
            # TODO: check to make sure we're not clearing other custom styles
            win.setStyleSheet(styleContent)

        # also set font size outside of stylesheet so that it is applied to new windows
        f = self._app.font()
        if mainFontSize is not None:
            f.setPointSizeF(mainFontSize)
            self._app.setFont(f)

    def _onSessionMiscSettingsChanged(self, whatChanged: list[str] | None = None):
        if whatChanged is None or any(x in whatChanged for x in ('theme', 'mainFontSize')):
            self._refreshGUIAppearance()

    def _onAddonsAboutToChange(self, itemKeys: list[str],
                               attribKeys: list[str] | None = None):
        if self._session is not None:
            if len(self._session.addons) > 0:
                if attribKeys is None and any(itemKey in self._session.addons for itemKey in itemKeys):
                    raise NotImplementedError  # TODO
                    # TODO: unload any addons changed not present in new session
                else:
                    pass  # assume any other changes don't need to be handled here

    def _onAddonsChanged(self, itemKeys: list[str],
                         attribKeys: list[str] | None = None,
                         triggeredBySessionLoad: bool = False):

        if attribKeys is not None and all(attribKey.startswith('SessionAttrs.') for attribKey in attribKeys):
            # change was only within addon session attributes, no need to do anything here
            return

        needToUpdateEnabledPanels = False
        prevActiveViewKey = self.activeViewKey
        for addonKey in itemKeys:
            if addonKey not in self._session.addons:
                continue  # assume addon was already unloaded

            addon = self._session.addons[addonKey]

            if addon.needsToInstantiateExtras:
                addon.instantiateExtras(navigatorGUI=self, session=self._session)

            for panelKey, ACE_Panel in addon.MainViewPanels.items():
                Panel = ACE_Panel.Class
                if panelKey not in self._mainViewPanels:
                    logger.info(f'Loading addon {addonKey} main view panel {panelKey}')
                    self._addViewPanel(Panel(key=panelKey))  # don't set session until after setting active view below to support deferred panel initialization
                    self._activateView(prevActiveViewKey)  # prevented switching focus to most recent loaded main view panel
                    self._mainViewPanels[panelKey].session = self._session

                    if not triggeredBySessionLoad:
                        needToUpdateEnabledPanels = True

                else:
                    logger.info(f'Addon {addonKey} main view panel {panelKey} already loaded, not reloading')

        if needToUpdateEnabledPanels:
            self._updateEnabledPanels()

    def _onSessionClosed(self, prevSession: Session):
        logger.info('Closed session {}'.format(prevSession.filepath))
        self._session = None
        for pane in self._mainViewPanels.values():
            pane.session = None
        self._updateEnabledPanels()

    @property
    def session(self):
        return self._session

    @property
    def mainViewPanels(self):
        return self._mainViewPanels

    @property
    def manageSessionPanel(self) -> ManageSessionPanel:
        return self._mainViewPanels['Manage session']

    @property
    def mriPanel(self) -> MRIPanel:
        return self._mainViewPanels['Set MRI']

    @property
    def headModelPanel(self) -> HeadModelPanel:
        return self._mainViewPanels['Set head model']

    @property
    def planFiducialsPanel(self) -> FiducialsPanel:
        return self._mainViewPanels['Plan fiducials']

    @property
    def setTargetsPanel(self) -> TargetsPanel:
        return self._mainViewPanels['Set targets']

    @property
    def toolsPanel(self) -> ToolsPanel:
        return self._mainViewPanels['Tools']

    @property
    def subjectRegistrationPanel(self) -> SubjectRegistrationPanel:
        return self._mainViewPanels['Register']

    @property
    def cameraPanel(self) -> CameraPanel:
        return self._mainViewPanels['Camera']

    @property
    def navigatePanel(self) -> NavigatePanel:
        return self._mainViewPanels['Navigate']

    @property
    def digitizeLocationsPanel(self) -> DigitizeLocationsPanel:
        return self._mainViewPanels['Digitize']

    async def _loadAfterSetup(self, filepath):
        await asyncio.sleep(1.)
        logger.info(f'Loading session from {filepath}')
        self._mainViewPanels['Manage session'].loadSession(sesFilepath=filepath)
        logger.debug('Done loading session')

    def _updateEnabledPanels(self):

        for key, panel in self._mainViewPanels.items():
            panel.updateEnabled()

        # TODO: if we just disabled the only active view, change to a useful fallback view
        fallbackViews = ['Manage session', 'Set MRI', 'Set head model', 'Plan fiducials', 'Register']

    @property
    def activeViewKey(self) -> str | None:
        """
        Note: This doesn't account for "active" views is secondary windows, split views, etc.
        """
        try:
            return self.manageSessionPanel.dockWdgt.parent().parent().stack.currentWidget().name()
        except:
            # can happen if manage session is not in a tabbed container
            return None

    def _activateView(self, viewKey: str):

        self._mainViewPanels[viewKey].dockWdgt.raiseDock()

        logger.info('Switched to view "{}"'.format(viewKey))


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--sesFilepath', type=str, default=None)
    args = parser.parse_args()

    if args.sesFilepath is None:
        if False:  # TODO: debug, delete or set to False
            if True:
                #sesFilepath = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', '..', '..', 'data/sub-2003_ses-test4.rtnabs')
                # sesFilepath = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', '..', '..', 'data/sub-2003_ses-test8.rtnabsdir')
                # sesFilepath = r'G:\My Drive\KellerLab\RTNaBS\data\sub-2355\ses-230209\sub-2355_ses-230209_RTNaBS'
                sesFilepath = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', '..', '..', 'data/sub-2003_ses-test9.rtnabsdir')
                sesFilepath = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', '..', '..', 'data/sub-1_ses-demo.navinibsdir')
                sesFilepath = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', '..', '..', 'data/sub-1_ses-cobotDemo-simPos.navinibsdir')
            else:
                sesFilepath = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', '..', '..',
                                           'data/TestSession1.rtnabs')
            NavigatorGUI.createAndRun(sesFilepath=sesFilepath)
        else:
            NavigatorGUI.createAndRun()
    else:
        NavigatorGUI.createAndRun(sesFilepath=args.sesFilepath)


if __name__ == '__main__':
    main()



from __future__ import annotations

import attrs
import logging
import numpy as np
import typing as tp
from typing import ClassVar

from . import PlotViewLayer
from .PlotLayersGroup import PlotLayersGroup
from NaviNIBS.Navigator.TargetingCoordinator import ProjectionSpecification
from NaviNIBS.util.pyvista import setActorUserTransform
from NaviNIBS.util.Transforms import composeTransform


logger = logging.getLogger(__name__)


Transform = np.ndarray


@attrs.define(kw_only=True)
class TargetingPointLayer(PlotViewLayer):
    _type: ClassVar[str] = 'TargetingPoint'

    _orientation: str  # 'target' or 'coil'
    _depth: tp.Union[str, ProjectionSpecification]
    """
    if orientation=='coil', can be one of ['coil', 'skin', 'gm', 'target']; if orientation=='target', can be ['coil', 'skin', 'gm', 'target']
    Otherwise can be ProjectionSpecification instance to choose depth based on projecting to another orientation
    """

    _color: str = '#0000ff'
    _opacity: float = 0.5
    _radius: float = 5.

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._coordinator.sigCurrentTargetChanged.connect(lambda: self._queueRedraw(which='all'))
        self._coordinator.sigCurrentCoilPositionChanged.connect(lambda: self._queueRedraw(which=['updatePosition']))

    def _getCoord(self, orientation: str, depth: tp.Union[str, ProjectionSpecification]) -> tp.Optional[np.ndarray]:
        return self._coordinator.getTargetingCoord(orientation=orientation, depth=depth)

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str, ...]] = None):
        super()._redraw(which=which)

        if not isinstance(which, str):
            # assume parent call above triggered appropriate redraws
            return

        if which == 'all':
            which = ['initPoint']
            self._redraw(which=which)
            return

        elif which == 'initPoint':
            actorKey = self._getActorKey('point')

            self._actors[actorKey] = self._plotter.add_points(np.asarray([0., 0., 0.]),
                                                              color=self._color,
                                                              opacity=self._opacity,
                                                              point_size=self._radius*2,
                                                              reset_camera=False,
                                                              render_points_as_spheres=True,
                                                              name=actorKey)

            self._redraw('updatePosition')

        elif which == 'updatePosition':
            actorKey = self._getActorKey('point')
            actor = self._actors[actorKey]

            coord = self._getCoord(orientation=self._orientation, depth=self._depth)

            with self._plotter.allowNonblockingCalls():
                if coord is None:
                    # no valid pose available
                    if actor.GetVisibility():
                        actor.VisibilityOff()
                        self._plotter.render()
                else:
                    if not actor.GetVisibility():
                        actor.VisibilityOn()
                    setActorUserTransform(actor, composeTransform(np.eye(3), coord))
                    self._plotter.render()

        else:
            raise NotImplementedError('Unexpected redraw which: {}'.format(which))


@attrs.define
class TargetingTargetPointsLayer(PlotLayersGroup):
    _type: ClassVar[str] = 'TargetingTargetPoints'

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.addLayer(TargetingPointLayer, 'targetCoilPoint', color='#0000ff', orientation='target', depth='coil')
        self.addLayer(TargetingPointLayer, 'targetTargetPoint', color='#0000ff', orientation='target', depth='target')


@attrs.define
class TargetingCoilPointsLayer(PlotLayersGroup):
    _type: ClassVar[str] = 'TargetingCoilPoints'

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.addLayer(TargetingPointLayer, 'coilCoilPoint', color='#00ff00', orientation='coil', depth='coil', radius=8)
        self.addLayer(TargetingPointLayer, 'coilTargetPoint', color='#00ff00', orientation='coil', depth='target', radius=8)
        if False:  # TODO: debug, enable by default
            self.addLayer(TargetingPointLayer, 'coilProjectedTargetPoint', color='#00ff00', orientation='coil',
                          depth=ProjectionSpecification(toOrientation='target', toDepth='target', toShape='sphere'))





import asyncio
import atexit
import logging
import typing as tp

from . import exceptionToStr

logger = logging.getLogger(__name__)


def asyncioRunAndHandleExceptions(fn: tp.Callable[..., tp.Awaitable], *args, **kwargs):
    def handleExceptionInAsyncTask(loop: asyncio.AbstractEventLoop, context: dict):
        from pprint import pprint
        logger.error('Unhandled exception in async task, will stop loop. Exception context:\n%s' % (pprint(context),))
        loop.stop()

    if False:
        loop = asyncio.get_event_loop()
    else:
        loop = asyncio.new_event_loop()
    loop.set_exception_handler(handleExceptionInAsyncTask)
    try:
        loop.run_until_complete(fn(*args, **kwargs))
    except RuntimeError as e:
        if e.args[0] == 'Event loop stopped before Future completed.':
            logger.info('Event loop stopped early')
        else:
            logger.error('Exception: %s' % exceptionToStr(e))
            raise e
    except Exception as e:
        logger.error('Exception: %s' % exceptionToStr(e))
        raise e
    finally:
        logger.debug('Running any shutdown_asyncgens')
        try:
            loop.run_until_complete(loop.shutdown_asyncgens())
        except RuntimeError as e:
            if e.args[0] == 'Event loop is closed':
                logger.info('Event looped already closed')
            else:
                raise e
        else:
            logger.debug('Closing loop')
            loop.close()

        logger.debug('Running any atexits')
        atexit._run_exitfuncs()

        logger.debug('Done!')



from __future__ import annotations
import attrs
import logging
import typing as tp

from NaviNIBS.util.attrs import attrsAsDict
from NaviNIBS.util.Signaler import Signal

logger = logging.getLogger(__name__)


@attrs.define
class MiscSettings:
    _mainFontSize: float | None = None
    _theme: str = 'light'

    sigAttribsChanged: Signal[list[str] | None] = attrs.field(init=False, factory=Signal)
    """
    Includes list of keys of attributes that changed, or None if any/all should be assumed to have changed.
    """

    def __attrs_post_init__(self):
        pass

    def asDict(self):
        return attrsAsDict(self)

    @classmethod
    def fromDict(cls, d: tp.Dict[str, tp.Any]):
        """
        Create a MiscSettings object from a dictionary.
        """
        return cls(**d)

    @property
    def mainFontSize(self):
        return self._mainFontSize

    @mainFontSize.setter
    def mainFontSize(self, value: float | None):
        if value == self._mainFontSize:
            return
        logger.info(f'Changing mainFontSize to {value}')
        self._mainFontSize = value
        self.sigAttribsChanged.emit(['mainFontSize'])

    @property
    def theme(self):
        return self._theme

    @theme.setter
    def theme(self, value: str):
        if value == self._theme:
            return
        logger.info(f'Changing theme to {value}')
        self._theme = value
        self.sigAttribsChanged.emit(['theme'])



from __future__ import annotations

import asyncio

import attrs
import logging
import multiprocessing as mp
import numpy as np
import qtawesome as qta
from qtpy import QtWidgets, QtGui
import typing as tp

from NaviNIBS.Devices.ToolPositionsServer import ToolPositionsServer
from NaviNIBS.Devices.ToolPositionsClient import ToolPositionsClient
from NaviNIBS.Devices.IGTLinkToolPositionsServer import IGTLinkToolPositionsServer
from NaviNIBS.Navigator.Model.Session import Session, SubjectTracker
from NaviNIBS.Navigator.GUI.ViewPanels.MainViewPanelWithDockWidgets import MainViewPanelWithDockWidgets
from NaviNIBS.Navigator.GUI.Widgets.TrackingStatusWidget import TrackingStatusWidget
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.pyvista import Actor, setActorUserTransform, RemotePlotterProxy
from NaviNIBS.util.GUI.Dock import Dock
from NaviNIBS.util.GUI.Icons import getIcon
from NaviNIBS.util.logging import getLogFilepath
from NaviNIBS.util.Transforms import invertTransform, concatenateTransforms, applyTransform
from NaviNIBS.util.GUI.QueuedRedrawMixin import QueuedRedrawMixin
from NaviNIBS.util.pyvista import DefaultBackgroundPlotter

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@attrs.define
class CameraObjectsView(QueuedRedrawMixin):
    _positionsClient: ToolPositionsClient
    _session: Session = attrs.field(repr=False)

    _plotter: DefaultBackgroundPlotter = attrs.field(init=False, repr=False)

    _actors: tp.Dict[str, tp.Optional[Actor]] = attrs.field(init=False, factory=dict, repr=False)

    finishedAsyncInit: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    def __attrs_post_init__(self):
        QueuedRedrawMixin.__attrs_post_init__(self)

        self._plotter = DefaultBackgroundPlotter()

        self._session.tools.sigItemsChanged.connect(lambda *args: self._queueRedraw('afterToolsChanged'))

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._finishInitialization_async))

    @property
    def plotter(self):
        return self._plotter

    @property
    def session(self):
        return self._session

    async def _finishInitialization_async(self):
        if isinstance(self._plotter, RemotePlotterProxy):
            await self._plotter.isReadyEvent.wait()

        if isinstance(self._plotter, RemotePlotterProxy):
            self._plotter.enable_depth_peeling(4)

        self._plotter.add_axes_at_origin(labels_off=True, line_width=4)

        self._positionsClient.sigLatestPositionsChanged.connect(lambda: self._queueRedraw('toolPositions'))

        self._redraw('all')

        self._autoOrientCamera()

        self.finishedAsyncInit.set()

    def _autoOrientCamera(self, distance: float = 1000):
        # if necessary info available, reorient camera view based on subject location
        if not self.session.subjectRegistration.isRegistered:
            logger.info('Not setting camera due lack of subject registration')
        else:
            subTracker = self.session.tools.subjectTracker
            if subTracker is None:
                logger.info('Not setting camera due lack of defined subject tracker')
            else:
                transf_subTrackToWorld = self._positionsClient.getLatestTransf(subTracker.key, None)
                if transf_subTrackToWorld is None:
                    logger.info('Not setting camera due to lack of valid subject tracker position')
                else:
                    transf_MRIToWorld = concatenateTransforms([
                        invertTransform(self.session.subjectRegistration.trackerToMRITransf),
                        transf_subTrackToWorld,
                    ])

                    # for now, assume MRI direction is consistent
                    # could alternatively look for "standard" fiducial names and define relative to those here
                    # or use MNI space
                    cameraPts_MRI = np.asarray([
                        [0, 0, 0],  # focal point
                        np.asarray([-0.8, 1, 0.3]) * distance,  # position
                        [0, 0, 1]])  # up

                    cameraPts_world = applyTransform(transf_MRIToWorld, cameraPts_MRI)

                    plotter = self._plotter
                    with plotter.allowNonblockingCalls():
                        plotter.camera.focal_point = cameraPts_world[0, :]
                        plotter.camera.position = cameraPts_world[1, :]
                        plotter.camera.up = cameraPts_world[2, :] - cameraPts_world[1, :]


    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str]] = None, **kwargs):
        super()._redraw(which=which, **kwargs)

        if isinstance(self._plotter, RemotePlotterProxy) and not self._plotter.isReadyEvent.is_set():
            # remote plotter not ready yet
            return

        if which is None:
            which = 'all'
            self._redraw(which=which, **kwargs)
            return

        if not isinstance(which, str):
            for subWhich in which:
                self._redraw(which=subWhich, **kwargs)
            return

        if which == 'all':
            self._redraw(which=['toolPositions'])
            return

        if which == 'toolPositions':
            doResetCamera = False

            for key, tool in self.session.tools.items():
                actorKeysForTool = [key + '_tracker', key + '_tool']
                if isinstance(tool, SubjectTracker):
                    actorKeysForTool.append(key + '_subject')

                if not tool.isActive or self._positionsClient.getLatestTransf(tool.trackerKey, None) is None:
                    # no valid position available
                    with self._plotter.allowNonblockingCalls():
                        for actorKey in actorKeysForTool:
                            if actorKey in self._actors and self._actors[actorKey].GetVisibility():
                                self._actors[actorKey].VisibilityOff()
                    continue

                for actorKey in actorKeysForTool:
                    logger.debug(f'actorKey: {actorKey}')

                    doShow = False
                    for toolOrTracker in ('tracker', 'tool'):
                        if actorKey == (key + '_' + toolOrTracker):
                            if getattr(tool, 'doRender' + toolOrTracker.capitalize()) is False:
                                doShow = False
                            else:
                                if getattr(tool, toolOrTracker + 'StlFilepath') is not None:
                                    if toolOrTracker == 'tool':
                                        if tool.toolToTrackerTransf is None:
                                            toolOrTrackerStlToTrackerTransf = None
                                        else:
                                            toolOrTrackerStlToTrackerTransf = tool.toolToTrackerTransf @ tool.toolStlToToolTransf
                                    elif toolOrTracker == 'tracker':
                                        toolOrTrackerStlToTrackerTransf = tool.trackerStlToTrackerTransf
                                    else:
                                        raise NotImplementedError()
                                    if toolOrTrackerStlToTrackerTransf is not None:
                                        doShow = True

                                        if actorKey not in self._actors:
                                            # initialize graphic
                                            mesh = getattr(tool, toolOrTracker + 'Surf')
                                            meshColor = tool.trackerColor if toolOrTracker == 'tracker' else tool.toolColor
                                            meshOpacity = tool.trackerOpacity if toolOrTracker == 'tracker' else tool.toolOpacity

                                            self._actors[actorKey] = self._plotter.addMesh(mesh=mesh,
                                                                                           color=meshColor,
                                                                                           defaultMeshColor='#444444',
                                                                                           opacity=1.0 if meshOpacity is None else meshOpacity,
                                                                                           name=actorKey)

                                            doResetCamera = True

                                    if doShow:
                                        with self._plotter.allowNonblockingCalls():
                                            # apply transform to existing actor
                                            setActorUserTransform(self._actors[actorKey],
                                                                  concatenateTransforms([
                                                                      toolOrTrackerStlToTrackerTransf,
                                                                      self._positionsClient.getLatestTransf(tool.trackerKey)
                                                                  ]))
                                            self._plotter.render()
                                else:
                                    # TODO: show some generic graphic to indicate tool position, even when we don't have an stl for the tool
                                    doShow = False

                    if isinstance(tool, SubjectTracker) and actorKey == tool.key + '_subject':
                        if self.session.subjectRegistration.trackerToMRITransf is not None and self.session.headModel.skinSurf is not None:
                            doShow = True
                            if actorKey not in self._actors:
                                self._actors[actorKey] = self._plotter.add_mesh(mesh=self.session.headModel.skinSurf,
                                                                                color='#d9a5b2',
                                                                                # opacity=0.8,
                                                                                name=actorKey)
                                doResetCamera = True

                            with self._plotter.allowNonblockingCalls():
                                setActorUserTransform(self._actors[actorKey],
                                                      self._positionsClient.getLatestTransf(tool.trackerKey) @ invertTransform(self.session.subjectRegistration.trackerToMRITransf))
                                self._plotter.render()

                    if actorKey in self._actors:
                        with self._plotter.allowNonblockingCalls():
                            if doShow and not self._actors[actorKey].GetVisibility():
                                self._actors[actorKey].VisibilityOn()
                                self._plotter.render()
                            elif not doShow and self._actors[actorKey].GetVisibility():
                                self._actors[actorKey].VisibilityOff()
                                self._plotter.render()

            if doResetCamera:
                if False:
                    with self._plotter.allowNonblockingCalls():
                        self._plotter.reset_camera()

        elif which == 'afterToolsChanged':
            # remove any tool actors, to be regenerated later
            didRemove = False
            for key, tool in self.session.tools.items():
                actorKeysForTool = [key + '_tracker', key + '_tool']
                with self._plotter.allowNonblockingCalls():
                    for actorKey in actorKeysForTool:
                        if actorKey in self._actors:
                            self._plotter.remove_actor(self._actors.pop(actorKey))
                            didRemove = True

            if didRemove:
                self._redraw('toolPositions')

        else:
            raise NotImplementedError


@attrs.define
class CameraPanel(MainViewPanelWithDockWidgets):
    """
    For now, assume this will always be connecting to an NDI Polaris camera with PyIGTLink.

    In the future, can update to have a more device-agnostic base class that is subclassed for specific localization systems
    """
    _key: str = 'Camera'
    _autoLayout: str | bool = True
    """
    Set to False to disable auto layout 
    """

    _icon: QtGui.QIcon = attrs.field(init=False, factory=lambda: getIcon('mdi6.cctv'))

    _trackingStatusWdgt: TrackingStatusWidget = attrs.field(init=False)
    _trackingStatusDock: Dock = attrs.field(init=False)

    _positionsServerProc: tp.Optional[mp.Process] = attrs.field(init=False, default=None)
    _positionsClient: ToolPositionsClient = attrs.field(init=False)

    _serverTypeComboBox: QtWidgets.QComboBox = attrs.field(init=False)
    _serverAddressEdit: QtWidgets.QLineEdit = attrs.field(init=False)
    _serverStartStopBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _serverGUIContainer: QtWidgets.QGroupBox = attrs.field(init=False)
    _serverGUIDock: Dock = attrs.field(init=False)

    _mainCameraView: CameraObjectsView = attrs.field(init=False)
    _mainCameraDock: Dock = attrs.field(init=False)

    finishedAsyncInit: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self.dockWdgt.sigResized.connect(self._onResized)

    def canBeEnabled(self) -> tuple[bool, str | None]:
        if self.session is None:
            return False, 'No session set'
        return True, None

    def _finishInitialization(self):
        super()._finishInitialization()

        self._trackingStatusWdgt = TrackingStatusWidget(session=self._session, wdgt=QtWidgets.QWidget())
        dock, _ = self._createDockWidget(
            title='Tracking status',
            widget=self._trackingStatusWdgt.wdgt)
        dock.setStretch(.1, 1)
        dock.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        self._wdgt.addDock(dock, position='left')
        self._trackingStatusDock = dock

        dock, container = self._createDockWidget(
            title='Camera connection',
            layout=QtWidgets.QVBoxLayout(),
        )
        dock.setStretch(.1, 1)
        dock.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        self._wdgt.addDock(dock, position='bottom')
        self._serverGUIDock = dock
        self._serverGUIContainer = container

        # TODO: add GUI controls for configuring, launching, stopping Plus Server
        # for now, assume plus server is launched separately with appropriate tool configs

        subContainer = QtWidgets.QGroupBox('Tool positions server')
        formLayout = QtWidgets.QFormLayout()
        subContainer.setLayout(formLayout)
        container.layout().addWidget(subContainer)

        self._serverTypeComboBox = QtWidgets.QComboBox()
        self._serverTypeComboBox.addItems(['IGTLink', 'Generic'])
        formLayout.addRow('Server type', self._serverTypeComboBox)

        self._serverAddressEdit = QtWidgets.QLineEdit()
        formLayout.addRow('Server addr', self._serverAddressEdit)
        self._serverAddressEdit.textChanged.connect(self._onServerAddressTextChanged)

        btn = QtWidgets.QPushButton('Start server' if self._positionsServerProc is not None else 'Stop server')
        btn.clicked.connect(self._onStartStopServerClicked)
        subContainer.layout().addWidget(btn)
        self._serverStartStopBtn = btn

        container.layout().addStretch()

        self._positionsClient = ToolPositionsClient()
        self._positionsClient.sigIsConnectedChanged.connect(self._onClientIsConnectedChanged)

        self._mainCameraView = CameraObjectsView(
            positionsClient=self._positionsClient,
            session=self.session
        )

        # TODO: create other camera views as defined in session config

        dock, _ = self._createDockWidget(
            title='Tracked objects',
            widget=self._mainCameraView.plotter)
        dock.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        dock.setStretch(2, 10)
        self._wdgt.addDock(dock, position='right')
        self._mainCameraDock = dock

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._finishInitialization_async))

    async def _finishInitialization_async(self):

        if self.session is not None:
            self._onPanelInitializedAndSessionSet()

        self._onClientIsConnectedChanged()

        await asyncio.sleep(0.5)
        self._updateAutoLayout()  # run after delay to allow others to resize first

        await self._mainCameraView.finishedAsyncInit.wait()

        self.finishedAsyncInit.set()

    def _onSessionSet(self):
        super()._onSessionSet()

        if self._positionsServerProc is not None:
            # kill previous server
            self._stopPositionsServer()

        if self.session.tools.positionsServerInfo.doAutostart:
            self._startPositionsServer()

        if self._hasInitialized:
            self._onPanelInitializedAndSessionSet()

    def _onPanelInitializedAndSessionSet(self):

        assert self._mainCameraView.session is self.session
        # TODO: assert for other views too

        self._trackingStatusWdgt.session = self.session

        info = self.session.tools.positionsServerInfo
        if info.type is None:
            self._serverTypeComboBox.setCurrentIndex(-1)
        else:
            self._serverTypeComboBox.setCurrentText(info.type)
        self._serverAddressEdit.setText(f'{info.hostname}:{info.pubPort},{info.cmdPort}')

    def _startPositionsServer(self):
        logger.info('Starting Positions server process')
        match(self.session.tools.positionsServerInfo.type):
            case 'IGTLink':
                Server = IGTLinkToolPositionsServer
            case 'Generic':
                Server = ToolPositionsServer
            case _:
                raise NotImplementedError(f'Unexpected positionsServerInfo type: {self.session.tools.positionsServerInfo.type}')

        kwargs = self.session.tools.positionsServerInfo.initKwargs.copy()
        kwargs['logFilepath'] = getLogFilepath(self.session)
        self._positionsServerProc = mp.Process(
            target=Server.createAndRun,
            daemon=True,
            kwargs=kwargs)
        self._positionsServerProc.start()
        if self._hasInitialized:
            self._serverStartStopBtn.setText('Stop server')

    def _stopPositionsServer(self):
        logger.info('Stopping Positions server process')
        self._positionsServerProc.kill()
        self._positionsServerProc = None
        if self._hasInitialized:
            self._serverStartStopBtn.setText('Start server')

    def _onStartStopServerClicked(self, checked: bool):
        if self._positionsServerProc is None:
            # start server
            self._startPositionsServer()
        else:
            # stop server
            self._stopPositionsServer()

    def _onServerAddressTextChanged(self, newVal: str):
        pass  # TODO: if changed, apply edits to ToolPositionsServerInfo

    def _onClientIsConnectedChanged(self):
        if not self._hasInitialized and not self.isInitializing:
            return

        if self._positionsClient.isConnected:
            self._serverStartStopBtn.setText('Stop server')

            serverType = self._positionsClient.getServerType()

            # TODO: maybe pop up a warning dialog here if old type does not match new server type,
            #  since this will overwrite setting in session file
            self.session.tools.positionsServerInfo.type = serverType
        else:
            self._serverStartStopBtn.setText('Start server')

    def _onLatestPositionsChanged(self):
        if not self._hasInitialized and not self.isInitializing:
            return

    def _onResized(self):
        self._updateAutoLayout()

    def _updateAutoLayout(self):
        if not self._hasInitialized:
            # not ready for auto layout yet
            return

        if isinstance(self._autoLayout, bool) and not self._autoLayout:
            # auto layout is disabled
            return

        if self.dockWdgt.width() > self.dockWdgt.height():
            newAutoLayout = 'horizontal'
        elif self.dockWdgt.width() > 800 and False:
            newAutoLayout = 'tallbig'
        else:
            newAutoLayout = 'tallsmall'

        if newAutoLayout == self._autoLayout:
            return

        self._autoLayout = newAutoLayout

        logger.info('Auto rearranging layout for panel size')

        match newAutoLayout:
            case 'horizontal':
                self._wdgt.moveDock(self._mainCameraDock, 'right', self._trackingStatusDock)
                self._wdgt.moveDock(self._serverGUIDock, 'bottom', self._trackingStatusDock)
                # TODO: maybe set stretches
            case 'tallbig':
                self._wdgt.moveDock(self._mainCameraDock, 'bottom', self._trackingStatusDock)
                self._wdgt.moveDock(self._serverGUIDock, 'right', self._trackingStatusDock)
            case 'tallsmall':
                self._wdgt.moveDock(self._mainCameraDock, 'bottom', self._trackingStatusDock)
                self._wdgt.moveDock(self._serverGUIDock, 'below', self._mainCameraDock)

            case _:
                raise NotImplementedError



    def close(self):
        if self._positionsServerProc is not None:
            self._stopPositionsServer()
        super().close()



from __future__ import annotations
import attrs


@attrs.define(frozen=True)
class ActorRef:
    """
    vtk Actors can't be pickled, so we use this class to pass to remote processes,
    and use ActorManager to keep track of which ref refers to which actor locally.
    """
    actorID: str





from __future__ import annotations

import asyncio
from datetime import datetime

import attrs
import json
import logging
import numpy as np
import os
import pathlib
import pyvista as pv
import pyvistaqt as pvqt
import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore
import shutil
import time
import typing as tp

from . import MainViewPanel
from NaviNIBS.Devices.ToolPositionsClient import ToolPositionsClient, TimestampedToolPosition
from NaviNIBS.Navigator.GUI.EditWindows.CoilCalibrationWindow import CoilCalibrationWithPlateWindow
from NaviNIBS.Navigator.GUI.EditWindows.PointerCalibrationWindow import PointerCalibrationWindow
from NaviNIBS.Navigator.GUI.Widgets.TrackingStatusWidget import TrackingStatusWidget
from NaviNIBS.Navigator.GUI.Widgets.CollectionTableWidget import ToolsTableWidget
from NaviNIBS.Navigator.Model.Session import Session, Tools, Tool, CoilTool, Pointer
from NaviNIBS.util import makeStrUnique
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.pyvista import setActorUserTransform, Actor, RemotePlotterProxy
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.Transforms import transformToString, stringToTransform, concatenateTransforms, invertTransform
from NaviNIBS.util.GUI.Icons import getIcon
from NaviNIBS.util.GUI.QFileSelectWidget import QFileSelectWidget
from NaviNIBS.util.GUI.QLineEdit import QLineEditWithValidationFeedback
from NaviNIBS.util.GUI.QTableWidgetDragRows import QTableWidgetDragRows
from NaviNIBS.util.GUI.QValidators import OptionalTransformValidator
from NaviNIBS.util.pyvista import DefaultBackgroundPlotter
from NaviNIBS.util.pyvista.plotting import BackgroundPlotter


logger = logging.getLogger(__name__)


@attrs.define
class ToolWidget:
    _tool: Tool
    _session: Session = attrs.field(repr=False)  # only used for cross-tool references like coil calibration

    _wdgt: QtWidgets.QWidget = attrs.field(init=False)
    _formLayout: QtWidgets.QFormLayout = attrs.field(init=False)
    _key: QtWidgets.QLineEdit = attrs.field(init=False)
    _trackerKey: QtWidgets.QLineEdit = attrs.field(init=False)
    _label: QtWidgets.QLineEdit = attrs.field(init=False)
    _usedFor: QtWidgets.QComboBox = attrs.field(init=False)
    _isActive: QtWidgets.QCheckBox = attrs.field(init=False)
    _romFilepath: QFileSelectWidget = attrs.field(init=False)
    _trackerStlFilepath: QFileSelectWidget = attrs.field(init=False)
    _toolStlFilepath: QFileSelectWidget = attrs.field(init=False)
    _toolToTrackerTransf: QtWidgets.QLineEdit = attrs.field(init=False)
    _toolStlToToolTransf: QtWidgets.QLineEdit = attrs.field(init=False)
    _trackerStlToTrackerTransf: QtWidgets.QLineEdit = attrs.field(init=False)
    _toolSpacePlotter: DefaultBackgroundPlotter = attrs.field(init=False)
    _trackerSpacePlotter: DefaultBackgroundPlotter = attrs.field(init=False)

    _asyncInitTask: asyncio.Task = attrs.field(init=False)
    finishedAsyncInit: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    _toolSpaceActors: dict[str, Actor] = attrs.field(init=False, factory=dict)
    _trackerSpaceActors: dict[str, Actor] = attrs.field(init=False, factory=dict)

    def __attrs_post_init__(self):
        self._wdgt = QtWidgets.QGroupBox('Selected tool: {}'.format(self._tool.key))
        self._wdgt.setLayout(QtWidgets.QVBoxLayout())

        formContainer = QtWidgets.QWidget()
        self._formLayout = QtWidgets.QFormLayout()
        formContainer.setLayout(self._formLayout)
        self._wdgt.layout().addWidget(formContainer)

        self._tool.sigItemChanged.connect(self._onToolChanged)

        self._key = QtWidgets.QLineEdit()
        self._key.editingFinished.connect(self._onKeyEdited)
        formContainer.layout().addRow('Key', self._key)

        self._trackerKey = QtWidgets.QLineEdit()
        self._trackerKey.editingFinished.connect(self._onTrackerKeyEdited)
        formContainer.layout().addRow('TrackerKey', self._trackerKey)

        self._label = QtWidgets.QLineEdit()
        self._label.editingFinished.connect(self._onLabelEdited)
        formContainer.layout().addRow('Label', self._label)

        self._usedFor = QtWidgets.QComboBox()
        self._usedFor.insertItems(0, ['coil', 'subject', 'pointer', 'calibration', 'visualization'])
        if len(self._tool.usedFor) > 0:
            index = self._usedFor.findText(self._tool.usedFor)
            assert index != -1, 'Unexpected tool type: {}'.format(self._tool.usedFor)
        else:
            index = -1
        self._usedFor.setCurrentIndex(index)
        self._usedFor.currentIndexChanged.connect(lambda index: self._onUsedForEdited())
        formContainer.layout().addRow('Type', self._usedFor)

        self._isActive = QtWidgets.QCheckBox('')
        self._isActive.setChecked(self._tool.isActive)
        self._isActive.stateChanged.connect(lambda state: self._onIsActiveEdited())
        formContainer.layout().addRow('Is active', self._isActive)

        self._romFilepath = QFileSelectWidget(
            browseMode='getOpenFilename',
            filepath=self._tool.romFilepath,
            showRelativeTo=self._tool.filepathsRelTo,
            showRelativePrefix=self._tool.filepathsRelToKey,
            extFilters='ROM (*.rom)',
            browseCaption='Choose tracker definition file',
        )
        self._romFilepath.sigFilepathChanged.connect(lambda filepath: self._onRomFilepathEdited())
        formContainer.layout().addRow('ROM filepath', self._romFilepath)

        self._trackerStlFilepath = QFileSelectWidget(
            browseMode='getOpenFilename',
            filepath=self._tool.trackerStlFilepath,
            showRelativeTo=self._tool.filepathsRelTo,
            showRelativePrefix=self._tool.filepathsRelToKey,
            extFilters='STL (*.stl); PLY (*.ply)',
            browseCaption='Choose 3D model for tracker visualization'
        )
        self._trackerStlFilepath.sigFilepathChanged.connect(lambda filepath: self._onTrackerStlFilepathEdited())
        formContainer.layout().addRow('Tracker STL filepath', self._trackerStlFilepath)

        self._toolStlFilepath = QFileSelectWidget(
            browseMode='getOpenFilename',
            filepath=self._tool.toolStlFilepath,
            showRelativeTo=self._tool.filepathsRelTo,
            showRelativePrefix=self._tool.filepathsRelToKey,
            extFilters='STL (*.stl); PLY (*.ply)',
            browseCaption='Choose 3D model for tool visualization'
        )
        self._toolStlFilepath.sigFilepathChanged.connect(lambda filepath: self._onToolStlFilepathEdited())
        formContainer.layout().addRow('Tool STL filepath', self._toolStlFilepath)

        self._toolStlToToolTransf = QLineEditWithValidationFeedback(self._transfToStr(self._tool.toolStlToToolTransf))
        self._toolStlToToolTransf.setValidator(OptionalTransformValidator())
        self._toolStlToToolTransf.editingFinished.connect(self._onToolStlToToolTransfEdited)
        formContainer.layout().addRow('Tool STL to tool transf', self._toolStlToToolTransf)

        self._trackerStlToTrackerTransf = QLineEditWithValidationFeedback(self._transfToStr(self._tool.trackerStlToTrackerTransf))
        self._trackerStlToTrackerTransf.setValidator(OptionalTransformValidator())
        self._trackerStlToTrackerTransf.editingFinished.connect(self._onTrackerStlToTrackerTransfEdited)
        formContainer.layout().addRow('Tracker STL to tracker transf', self._trackerStlToTrackerTransf)

        self._toolToTrackerTransf = QLineEditWithValidationFeedback(self._transfToStr(self._tool.toolToTrackerTransf))
        self._toolToTrackerTransf.setValidator(OptionalTransformValidator())
        self._toolToTrackerTransf.editingFinished.connect(self._onToolToTrackerTransfEdited)
        formContainer.layout().addRow('Tool to tracker transf', self._toolToTrackerTransf)

        plotContainer = QtWidgets.QWidget()
        plotContainer.setLayout(QtWidgets.QHBoxLayout())
        self._wdgt.layout().addWidget(plotContainer)

        self._toolSpacePlotter = DefaultBackgroundPlotter(parent=self._wdgt)
        # TODO: make sure these plotters (when RemotePlotterProxies) are appropriately terminated when switching between tools
        # TODO: ideally actually repurpose previous RemotePlotter when switching between tools to avoid unnecessary process init times

        plotterContainer = QtWidgets.QGroupBox('Tool space')
        plotterContainer.setLayout(QtWidgets.QVBoxLayout())
        plotterContainer.layout().addWidget(self._toolSpacePlotter)
        plotContainer.layout().addWidget(plotterContainer)

        self._trackerSpacePlotter = DefaultBackgroundPlotter(parent=self._wdgt)

        plotterContainer = QtWidgets.QGroupBox('Tracker space')
        plotterContainer.setLayout(QtWidgets.QVBoxLayout())
        plotterContainer.layout().addWidget(self._trackerSpacePlotter)
        plotContainer.layout().addWidget(plotterContainer)

        self._asyncInitTask = asyncio.create_task(asyncTryAndLogExceptionOnError(self._finishInitialization_async))

        QtCore.QTimer.singleShot(0, lambda: self._onToolChanged(self._tool.key, attribsChanged=None))

        self.redraw()

    async def _finishInitialization_async(self):
        if isinstance(self._toolSpacePlotter, RemotePlotterProxy):
            await self._toolSpacePlotter.isReadyEvent.wait()

        if isinstance(self._trackerSpacePlotter, RemotePlotterProxy):
            await self._trackerSpacePlotter.isReadyEvent.wait()

        for plotter in (self._toolSpacePlotter, self._trackerSpacePlotter):
            with plotter.allowNonblockingCalls():
                plotter.enable_parallel_projection()
                plotter.enable_depth_peeling(4)

        self.finishedAsyncInit.set()

        self.redraw()

    def redraw(self, whatToRedraw: tp.Iterable[str] | None = None):

        if not self.finishedAsyncInit.is_set():
            return

        if whatToRedraw is None or 'tool' in whatToRedraw:
            actorKey = 'tool'
            if actorKey in self._toolSpaceActors:
                self._toolSpacePlotter.remove_actor(self._toolSpaceActors.pop(actorKey))
            if actorKey in self._trackerSpaceActors:
                self._trackerSpacePlotter.remove_actor(self._trackerSpaceActors.pop(actorKey))

            defaultGridKwargs = dict(
                bold=False,
            )

            if self._tool.toolSurf is not None:
                meshColor = self._tool.toolColor
                meshColor_tool = meshColor
                actor = self._toolSpacePlotter.addMesh(
                    name=actorKey,
                    mesh=self._tool.toolSurf,
                    color=meshColor,
                    defaultMeshColor='#444444',
                    opacity=0.8
                )
                self._toolSpaceActors[actorKey] = actor
                setActorUserTransform(actor, self._tool.toolStlToToolTransf)
                self._toolSpacePlotter.showGrid(
                    color=self._toolSpacePlotter.palette().color(QtGui.QPalette.Text).name(),
                    **defaultGridKwargs)

            if self._tool.toolToTrackerTransf is not None and self._tool.toolSurf is not None:
                actor = self._trackerSpacePlotter.addMesh(
                    mesh=self._tool.toolSurf,
                    color=meshColor_tool,  # noqa
                    defaultMeshColor='#444444',
                    opacity=0.8,
                    name=actorKey
                )
                self._trackerSpaceActors[actorKey] = actor
                setActorUserTransform(actor, self._tool.toolToTrackerTransf @ self._tool.toolStlToToolTransf)
                self._trackerSpacePlotter.showGrid(
                    color=self._trackerSpacePlotter.palette().color(QtGui.QPalette.Text).name(),
                    **defaultGridKwargs)

            self._toolSpacePlotter.reset_camera()

        if whatToRedraw is None or 'tracker' in whatToRedraw:
            actorKey = 'tracker'
            if actorKey in self._toolSpaceActors:
                self._toolSpacePlotter.remove_actor(self._toolSpaceActors.pop(actorKey))
            if actorKey in self._trackerSpaceActors:
                self._trackerSpacePlotter.remove_actor(self._trackerSpaceActors.pop(actorKey))

            if self._tool.trackerStlToTrackerTransf is not None and self._tool.trackerSurf is not None:

                defaultGridKwargs = dict(
                    bold=False,
                )

                meshColor = self._tool.trackerColor
                actor = self._trackerSpacePlotter.addMesh(
                    mesh=self._tool.trackerSurf,
                    color=meshColor,
                    defaultMeshColor='#444444',
                    opacity=0.8,
                    name=actorKey
                )
                self._trackerSpaceActors[actorKey] = actor
                setActorUserTransform(actor, self._tool.trackerStlToTrackerTransf)
                self._trackerSpacePlotter.showGrid(
                    color=self._trackerSpacePlotter.palette().color(QtGui.QPalette.Text).name(),
                    **defaultGridKwargs)

                if self._tool.toolToTrackerTransf is not None:
                    actor = self._toolSpacePlotter.addMesh(
                        mesh=self._tool.trackerSurf,
                        color=meshColor,
                        defaultMeshColor='#444444',
                        opacity=0.8,
                        name=actorKey
                    )
                    self._toolSpaceActors[actorKey] = actor
                    setActorUserTransform(actor, concatenateTransforms([self._tool.trackerStlToTrackerTransf, invertTransform(self._tool.toolToTrackerTransf)]))
                    self._toolSpacePlotter.showGrid(
                        color=self._toolSpacePlotter.palette().color(QtGui.QPalette.Text).name(),
                        **defaultGridKwargs)

            self._trackerSpacePlotter.reset_camera()

    @property
    def wdgt(self):
        return self._wdgt

    def _onKeyEdited(self):
        self._tool.key = self._key.text()

    def _onTrackerKeyEdited(self):
        self._tool.trackerKey = self._trackerKey.text()

    def _onLabelEdited(self):
        newLabel = self._label.text().strip()
        if len(newLabel) == 0:
            newLabel = None
        self._tool.label = self._tool.label = newLabel

    def _onUsedForEdited(self):
        self._tool.usedFor = self._usedFor.currentText()

    def _onIsActiveEdited(self):
        self._tool.isActive = self._isActive.isChecked()

    def _onRomFilepathEdited(self):
        self._tool.romFilepath = self._romFilepath.filepath

    def _onTrackerStlFilepathEdited(self):
        self._tool.trackerStlFilepath = self._trackerStlFilepath.filepath

    def _onToolStlFilepathEdited(self):
        self._tool.toolStlFilepath = self._toolStlFilepath.filepath

    def _onToolStlToToolTransfEdited(self):
        newTransf = self._strToTransf(self._toolStlToToolTransf.text())
        if self._transfToStr(newTransf) == self._transfToStr(self._tool.toolStlToToolTransf):
            # no change
            return
        logger.info('User edited {} toolStlToToolTransf: {}'.format(self._tool.key, newTransf))
        self._tool.toolStlToToolTransf = newTransf

    def _onTrackerStlToTrackerTransfEdited(self):
        newTransf = self._strToTransf(self._trackerStlToTrackerTransf.text())
        if self._transfToStr(newTransf) == self._transfToStr(self._tool.trackerStlToTrackerTransf):
            # no change
            return
        logger.info('User edited {} trackerStlToTrackerTransf: {}'.format(self._tool.key, newTransf))
        self._tool.trackerStlToTrackerTransf = newTransf

    def _onToolToTrackerTransfEdited(self):
        newTransf = self._strToTransf(self._toolToTrackerTransf.text())
        if self._transfToStr(newTransf) == self._transfToStr(self._tool.toolToTrackerTransf):
            # no change
            return
        logger.info('User edited {} toolToTrackerTransf: {}'.format(self._tool.key, newTransf))
        self._tool.toolToTrackerTransf = newTransf

    def _onToolChanged(self, toolKey: str, attribsChanged: list[str] | None = None):
        toRedraw = set()
        if attribsChanged is None or 'key' in attribsChanged:
            self._key.setText(self._tool.key)
            self._trackerKey.setPlaceholderText(self._tool.key)
            self._label.setPlaceholderText(self._tool.key)
        if attribsChanged is None or 'trackerKey' in attribsChanged:
              if self._tool.trackerKeyIsSet:
                  self._trackerKey.setText(self._tool.trackerKey)
              else:
                  self._trackerKey.setText('')
        if attribsChanged is None or 'label' in attribsChanged:
            if self._tool.labelIsSet:
                self._label.setText(self._tool.label)
            else:
                self._label.setText('')
        if attribsChanged is None or 'usedFor' in attribsChanged:
            self._usedFor.setCurrentIndex(self._usedFor.findText(self._tool.usedFor) if self._tool.usedFor is not None else -1)  # TODO: check for change in type that we can't handle without reinstantiating
        if attribsChanged is None or 'isActive' in attribsChanged:
            self._isActive.setChecked(self._tool.isActive)
        if attribsChanged is None or 'romFilepath' in attribsChanged:
            self._romFilepath.filepath = self._tool.romFilepath
        if attribsChanged is None or 'trackerStlFilepath' in attribsChanged:
            self._trackerStlFilepath.filepath = self._tool.trackerStlFilepath
            toRedraw.add('tracker')
        if attribsChanged is None or 'toolStlFilepath' in attribsChanged:
            self._toolStlFilepath.filepath = self._tool.toolStlFilepath
            toRedraw.add('tool')
        if attribsChanged is None or 'filepathsRelToKey' in attribsChanged:
            for wdgt in (self._romFilepath, self._trackerStlFilepath, self._toolStlFilepath):
                wdgt.showRelativePrefix = self._tool.filepathsRelToKey
        if attribsChanged is None or 'toolStlToToolTransf' in attribsChanged:
            self._toolStlToToolTransf.setText(self._transfToStr(self._tool.toolStlToToolTransf))
            toRedraw.add('tool')
        if attribsChanged is None or 'trackerStlToTrackerTransf' in attribsChanged:
            self._trackerStlToTrackerTransf.setText(self._transfToStr(self._tool.trackerStlToTrackerTransf))
            toRedraw.add('tracker')
        if attribsChanged is None or 'toolToTrackerTransf' in attribsChanged:
            self._toolToTrackerTransf.setText(self._transfToStr(self._tool.toolToTrackerTransf))
            toRedraw |= {'tracker', 'tool'}

        if attribsChanged is None:
            self.redraw()
        else:
            self.redraw(toRedraw)

    def close(self):
        if not self.finishedAsyncInit.is_set():
            self._asyncInitTask.cancel()
        self._tool.sigItemChanged.disconnect(self._onToolChanged)
        self._toolSpacePlotter.close()
        self._trackerSpacePlotter.close()
        self._wdgt.deleteLater()  # TODO: verify this is correct way to remove from layout and also delete children

    @staticmethod
    def _transfToStr(transf: tp.Optional[np.ndarray]) -> str:
        if transf is None:
            return ''
        else:
            return transformToString(transf, precision=6)

    @staticmethod
    def _strToTransf(inputStr: str) -> tp.Optional[np.ndarray]:
        if len(inputStr.strip()) == 0:
            return None
        else:
            return stringToTransform(inputStr)


@attrs.define
class CoilToolWidget(ToolWidget):
    _tool: CoilTool

    _calibrateCoilBtn: QtWidgets.QPushButton = attrs.field(init=False)

    _lastCalibratedAtLabel: QtWidgets.QLabel = attrs.field(init=False)

    _calibrationWindow: CoilCalibrationWithPlateWindow | None = attrs.field(init=False, default=None)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._lastCalibratedAtLabel = QtWidgets.QLabel('')
        self._formLayout.addRow('Last calibrated at', self._lastCalibratedAtLabel)
        self._updateLastCalibratedAt()

        btn = QtWidgets.QPushButton('Calibrate coil...')
        self._calibrateCoilBtn = btn
        self._formLayout.addRow('', btn)
        btn.clicked.connect(lambda _: self._calibrate())

    def _updateLastCalibratedAt(self):
        historyTimes = [datetime.strptime(x, '%y%m%d%H%M%S.%f') for x in self._tool.toolToTrackerTransfHistory.keys()]
        if len(historyTimes) == 0:
            self._lastCalibratedAtLabel.setText('Never')
        else:
            lastCalibratedAt = historyTimes[-1]  # assume last in list is most recent
            self._lastCalibratedAtLabel.setText(lastCalibratedAt.strftime('%Y-%m-%d %H:%M:%S'))

    def _calibrate(self):
        assert self._calibrationWindow is None
        self._calibrationWindow = CoilCalibrationWithPlateWindow(
            parent=self._wdgt,
            toolKeyToCalibrate=self._tool.key,
            session=self._session
        )
        self._calibrationWindow.sigFinished.connect(self._onCalibrationWindowFinished)
        self._calibrationWindow.show()

    def _onCalibrationWindowFinished(self, *args):
        assert self._calibrationWindow is not None
        self._calibrationWindow.sigFinished.disconnect(self._onCalibrationWindowFinished)
        self._calibrationWindow = None

    def _onToolChanged(self, toolKey: str, attribsChanged: list[str] | None = None):
        super()._onToolChanged(toolKey=toolKey, attribsChanged=attribsChanged)
        self._updateLastCalibratedAt()


@attrs.define
class PointerToolWidget(ToolWidget):
    _tool: Pointer

    _calibrationWindow: CoilCalibrationWithPlateWindow | PointerCalibrationWindow | None = attrs.field(init=False, default=None)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        btn = QtWidgets.QPushButton('Calibrate pointer by multiple endpoint samples...')
        self._formLayout.addRow('', btn)
        btn.clicked.connect(lambda _: self._calibrateByEndpoint())

        btn = QtWidgets.QPushButton('Calibrate pointer with calibration plate...')
        self._formLayout.addRow('', btn)
        btn.clicked.connect(lambda _: self._calibrateWithPlate())

    def _calibrateWithPlate(self):
        # TODO: add extra arg to specify that pointer will be rotated 90 deg (tangential to calibration plate instead of perpendicular)
        assert self._calibrationWindow is None
        self._calibrationWindow = CoilCalibrationWithPlateWindow(
            parent=self._wdgt,
            toolKeyToCalibrate=self._tool.key,
            session=self._session
        )
        self._calibrationWindow.sigFinished.connect(self._onCalibrationWindowFinished)
        self._calibrationWindow.show()

    def _calibrateByEndpoint(self):
        assert self._calibrationWindow is None
        self._calibrationWindow = PointerCalibrationWindow(
            parent=self._wdgt,
            toolKeyToCalibrate=self._tool.key,
            session=self._session
        )
        self._calibrationWindow.sigFinished.connect(self._onCalibrationWindowFinished)
        self._calibrationWindow.show()

    def _onCalibrationWindowFinished(self, *args):
        assert self._calibrationWindow is not None
        self._calibrationWindow.sigFinished.disconnect(self._onCalibrationWindowFinished)
        self._calibrationWindow = None


@attrs.define
class ToolsPanel(MainViewPanel):
    _key: str = 'Tools'
    _icon: QtGui.QIcon = attrs.field(init=False, factory=lambda: getIcon('mdi6.hammer-screwdriver'))
    _trackingStatusWdgt: TrackingStatusWidget = attrs.field(init=False)
    _tblWdgt: ToolsTableWidget = attrs.field(init=False)
    _toolWdgt: tp.Optional[ToolWidget] = attrs.field(init=False, default=None)
    _wdgts: tp.Dict[str, QtWidgets.QWidget] = attrs.field(init=False, factory=dict)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def canBeEnabled(self) -> tuple[bool, str | None]:
        if self.session is None:
            return False, 'No session set'
        return True, None

    def _finishInitialization(self):
        super()._finishInitialization()

        self._wdgt.setLayout(QtWidgets.QHBoxLayout())

        sidebar = QtWidgets.QWidget()
        sidebar.setLayout(QtWidgets.QVBoxLayout())
        self._wdgt.layout().addWidget(sidebar)
        self._wdgt.layout().setAlignment(sidebar, QtCore.Qt.AlignLeft)
        sidebar.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.MinimumExpanding)

        self._trackingStatusWdgt = TrackingStatusWidget(session=self.session, hideInactiveTools=False)
        sidebar.layout().addWidget(self._trackingStatusWdgt.wdgt)

        container = QtWidgets.QGroupBox('Tools')
        container.setLayout(QtWidgets.QVBoxLayout())
        sidebar.layout().addWidget(container)

        btnContainer = QtWidgets.QWidget()
        btnContainer.setLayout(QtWidgets.QGridLayout())
        container.layout().addWidget(btnContainer)

        btn = QtWidgets.QPushButton('Import tool settings from file...')
        btn.clicked.connect(self._onImportToolsBtnClicked)
        btnContainer.layout().addWidget(btn, 0, 0, 1, 3)

        btn = QtWidgets.QPushButton('Add')
        btn.clicked.connect(self._onAddBtnClicked)
        btnContainer.layout().addWidget(btn, 1, 0)

        btn = QtWidgets.QPushButton('Duplicate')
        btn.clicked.connect(self._onDuplicateBtnClicked)
        btnContainer.layout().addWidget(btn, 1, 1)

        btn = QtWidgets.QPushButton('Delete')
        btn.clicked.connect(self._onDeleteBtnClicked)
        btnContainer.layout().addWidget(btn, 1, 2)

        self._tblWdgt = ToolsTableWidget(session=self.session)
        self._tblWdgt.sigCurrentItemChanged.connect(lambda *args: self._updateSelectedToolWdgt())
        container.layout().addWidget(self._tblWdgt.wdgt)

        if self.session is not None:
            self._onPanelInitializedAndSessionSet()

        self._updateSelectedToolWdgt()

    def _onSessionSet(self):
        super()._onSessionSet()

        if any(tool.initialTrackerPose is not None for tool in self.session.tools.values()):
            asyncio.create_task(asyncTryAndLogExceptionOnError(self._recordInitialToolPoses))

        self._session.tools.sigItemsChanged.connect(self._onToolsChanged)

        if self._hasInitialized:
            self._onPanelInitializedAndSessionSet()

    def _onToolsChanged(self, toolKeys: list[str], changedAttribs: list[str] | None = None):
        if changedAttribs is None \
                or any(x in changedAttribs for x in (
                'initialTrackerPose',
                'initialTrackerPoseRelativeTo')):
            asyncio.create_task(asyncTryAndLogExceptionOnError(self._recordInitialToolPoses, toolKeys=toolKeys, timestamp=time.time()))

    async def _recordInitialToolPoses(self, toolKeys: list[str] | None = None,
                                      timestamp: float | None = None):

        if toolKeys is None:
            toolKeys = self.session.tools.keys()

        if timestamp is None:
            timestamp = time.time()

        positionsClient = ToolPositionsClient()
        for toolKey in toolKeys:
            try:
                tool = self.session.tools[toolKey]
            except KeyError as e:
                # tool presumably removed since task was started
                continue
            if tool.initialTrackerPose is not None:
                logger.info(f'Recording initial tracker pose for {toolKey}: {tool.initialTrackerPose}')
                await positionsClient.recordNewPosition_async(
                    tool.trackerKey,
                    position=TimestampedToolPosition(
                        time=timestamp,
                        transf=tool.initialTrackerPose,
                        relativeTo=tool.initialTrackerPoseRelativeTo
                    ))

    def _onPanelInitializedAndSessionSet(self):
        self._trackingStatusWdgt.session = self.session
        self._tblWdgt.session = self.session
        self._updateSelectedToolWdgt()

    def _onTblCurrentCellChanged(self, key: str):
        self._updateSelectedToolWdgt()

    def _onImportToolsBtnClicked(self, checked: bool):
        newFilepath, _ = QtWidgets.QFileDialog.getOpenFileName(self._wdgt,
                                                               'Select tools file to import',
                                                               os.path.dirname(self.session.filepath),
                                                               'json (*.json);; NaviNIBS (*.navinibs)')

        if len(newFilepath) == 0:
            logger.warning('Import cancelled')
            return

        self._importToolsFromFile(newFilepath=newFilepath)

    def _importToolsFromFile(self, newFilepath: str):
            self.session.mergeFromFile(filepath=newFilepath, sections=['tools'])

            self._tblWdgt.resizeColumnsToContents()  # resize immediately rather than waiting for delayed auto-resize

    def _updateSelectedToolWdgt(self):
        logger.debug('Updating selected tool widget')
        currentToolKey = self._tblWdgt.currentCollectionItemKey
        # TODO: if possible, only update specific fields rather than fully recreating widget
        if self._toolWdgt is not None:
            self._toolWdgt.close()
            self._toolWdgt = None

        if currentToolKey is None:
            return

        if isinstance(self.session.tools[currentToolKey], CoilTool):
            ToolWidgetCls = CoilToolWidget
        elif isinstance(self.session.tools[currentToolKey], Pointer):
            ToolWidgetCls = PointerToolWidget
        else:
            ToolWidgetCls = ToolWidget

        self._toolWdgt = ToolWidgetCls(tool=self.session.tools[currentToolKey], session=self.session)
        self._wdgt.layout().addWidget(self._toolWdgt.wdgt)

    def _onAddBtnClicked(self, checked: bool):
        logger.info('Add tool btn clicked')
        self.session.tools.addItemFromDict(dict(key=makeStrUnique('Tool', self.session.tools.keys()), usedFor='coil'))

    def _onDuplicateBtnClicked(self, checked: bool):
        logger.info('Duplicate tool btn clicked')
        toolDict = self._tblWdgt.currentCollectionItem.asDict().copy()
        toolDict['key'] = makeStrUnique(toolDict['key'], self.session.tools.keys())
        self.session.tools.addToolFromDict(toolDict)

    def _onDeleteBtnClicked(self, checked: bool):
        keysToDelete = [key for key in self._tblWdgt.selectedCollectionItemKeys if key is not None]
        if len(keysToDelete) > 0:
            self.session.tools.deleteItems(keysToDelete)














import asyncio
import attrs
import logging
import typing as tp

from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@attrs.define(slots=False, kw_only=True)
class QueuedRedrawMixin:
    """
    Convenience mixin for classes with a _redraw method, to support "queueing" redraws.
    Especially useful for queueing redraws spawned by Qt events but that we want to handle
    in async loop, and for reducing redundant redraws.
    """
    _redrawQueue: list[str | tuple[str, dict]] = attrs.field(init=False, factory=list)
    _redrawQueueModified: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    redrawQueueIsEmpty: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    def __attrs_post_init__(self):
        self.redrawQueueIsEmpty.set()
        asyncio.create_task(asyncTryAndLogExceptionOnError(self._loop_queuedRedraw))

    async def _loop_queuedRedraw(self):
        while True:
            await self._redrawQueueModified.wait()
            await asyncio.sleep(0.01)  # rate limit  # TODO: make this a parameter
            while len(self._redrawQueue) > 0:
                toRedraw = self._redrawQueue.pop(0)
                logger.debug(f'Dequeuing redraw for {self.__class__.__name__} {toRedraw}')
                if isinstance(toRedraw, str):
                    self._redraw(which=toRedraw)
                else:
                    # tuple of (which, kwargs)
                    assert len(toRedraw) == 2
                    self._redraw(which=toRedraw[0], **toRedraw[1])
            self._redrawQueueModified.clear()
            self.redrawQueueIsEmpty.set()
            await asyncio.sleep(0.01)  # rate limit  # TODO: make this a parameter

    def _queueRedraw(self, which: tp.Union[tp.Optional[str], tp.List[str]] = None, **kwargs):

        if False:  # TODO: debug, delete or set to False
            self._redraw(which=which, **kwargs)
            return

        if which is None:
            which = 'all'

        if not isinstance(which, str):
            for subWhich in which:
                self._queueRedraw(which=subWhich, **kwargs)
            return

        if which == 'all':
            self._redrawQueue.clear()
        else:
            if 'all' in self._redrawQueue:
                return
            if which in self._redrawQueue:
                return  # don't need to add duplicate to queue
                # (assumes order of redraws doesn't matter)

        if len(kwargs) == 0:
            queueKey = which
        else:
            queueKey = (which, kwargs)
        logger.debug(f'Queueing redraw for {self.__class__.__name__} {queueKey}')
        self._redrawQueue.append(queueKey)
        self.redrawQueueIsEmpty.clear()
        self._redrawQueueModified.set()

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str]] = None, **kwargs):
        if isinstance(which, str):
            # since we're redrawing now, can remove this from the queue
            if len(kwargs) == 0:
                queueKey = which
            else:
                queueKey = (which, kwargs)
            try:
                self._redrawQueue.remove(queueKey)
            except ValueError:
                pass
            else:
                if len(self._redrawQueue) == 0:
                    self.redrawQueueIsEmpty.set()
                self._redrawQueueModified.set()

        # subclass should handle the rest


import numpy as np
import pyvista as pv
import typing as tp

Actor = pv._vtk.vtkActor


def setActorUserTransform(actor: Actor, transf: np.ndarray):
    t = pv._vtk.vtkTransform()
    t.SetMatrix(pv.vtkmatrix_from_array(transf))
    actor.SetUserTransform(t)


# test if running on mac
import platform
isMac = platform.system() == 'Darwin'
if False or isMac:
    from NaviNIBS.util.pyvista.plotting import BackgroundPlotter, PrimaryLayeredPlotter, SecondaryLayeredPlotter
    DefaultBackgroundPlotter = BackgroundPlotter
    RemotePlotterProxy = type('__None')  # for callers to easily check if DefaultBackgroundPlotter is RemotePlotterProxy
    DefaultPrimaryLayeredPlotter = PrimaryLayeredPlotter
    DefaultSecondaryLayeredPlotter = SecondaryLayeredPlotter
else:
    # note: mac does not support remote plotting via qt window embeddeding
    from NaviNIBS.util.pyvista.RemotePlotting.RemotePlotterProxy import RemotePlotterProxy
    from NaviNIBS.util.pyvista.RemotePlotting.RemoteLayeredPlotterProxies import RemotePrimaryLayeredPlotterProxy, RemoteSecondaryLayeredPlotterProxy

    DefaultBackgroundPlotter = RemotePlotterProxy
    DefaultPrimaryLayeredPlotter = RemotePrimaryLayeredPlotterProxy
    DefaultSecondaryLayeredPlotter = RemoteSecondaryLayeredPlotterProxy


def getActorUserTransform(actor: Actor) -> np.ndarray:
    return pv.array_from_vtkmatrix(actor.GetUserTransform().GetMatrix())


def concatenateLineSegments(lineSegmentGroups: tp.Iterable[pv.PolyData]) -> pv.PolyData:
    poly = pv.PolyData()
    poly.points = np.vstack([group.points for group in lineSegmentGroups])
    cells = np.vstack([group.lines.reshape((-1, 3)) for group in lineSegmentGroups])
    offset_points = 0
    offset_lines = 0
    for iGroup, group in enumerate(lineSegmentGroups):
        lines_group = group.lines.reshape((-1, 3))
        cells[offset_lines:(offset_lines+lines_group.shape[0]), 1:] += offset_points
        offset_points += group.points.shape[0]
        offset_lines += lines_group.shape[0]

    poly.lines = cells
    return poly




from __future__ import annotations

import asyncio
import attrs
import logging
from qtpy import QtWidgets, QtGui, QtCore
import typing as tp

from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.Navigator.Model.DockWidgetLayouts import DockWidgetLayout
from NaviNIBS.Navigator.GUI.ViewPanels import MainViewPanel
from NaviNIBS.util import exceptionToStr
from NaviNIBS.util.GUI.Dock import Dock, DockArea
from NaviNIBS.util.Signaler import Signal

logger = logging.getLogger(__name__)


@attrs.define
class MainViewPanelWithDockWidgets(MainViewPanel):
    _wdgt: DockArea = attrs.field(init=False)

    _dockWidgets: dict[str, Dock] = attrs.field(init=False, factory=dict)
    _layoutMayHaveChanged: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    sigAboutToRestoreLayout: Signal = attrs.field(init=False, factory=Signal)
    sigRestoredLayout: Signal = attrs.field(init=False, factory=Signal)

    def __attrs_post_init__(self):
        self._wdgt = DockArea(affinities=[self._key])
        self._wdgt.setContentsMargins(2, 2, 2, 2)

        super().__attrs_post_init__()

    def _createDockWidget(self,
                          title: str,
                          widget: tp.Optional[QtWidgets.QWidget] = None,
                          layout: tp.Optional[QtWidgets.QLayout] = None):

        dock = Dock(name=self._key + title,
                   title=title,
                   closable=False,
                   affinities=[self._key])

        if widget is None:
            widget = QtWidgets.QWidget()
        if layout is not None:
            widget.setLayout(layout)
        dock.addWidget(widget)
        self._dockWidgets[title] = dock
        return dock, widget

    def restoreLayoutIfAvailable(self) -> bool:
        """
        Returns True if layout was restored, False otherwise.

        Can be called by subclass after delayed initialization to restore layout at that time.
        """
        if self.session is None:
            return False

        layout = self.session.dockWidgetLayouts.get(self._key, None)
        if layout is None or layout.state is None:
            return False

        assert layout.affinities == [self._key]

        logger.debug(f'About to restore layout for {self._key}')
        self.sigAboutToRestoreLayout.emit()
        try:
            self._wdgt.restoreState(layout.state)
        except Exception as e:
            logger.warning(f'Unable to restore layout: {exceptionToStr(e)}')
            return False

        logger.debug(f'Restored layout for {self._key}')
        self.sigRestoredLayout.emit()
        return True

    def saveLayout(self):
        assert self.session is not None

        layout = self.session.dockWidgetLayouts.get(self._key, None)
        if layout is None:
            layout = DockWidgetLayout(
                key=self._key,
                affinities=[self._key])
            self.session.dockWidgetLayouts.addItem(layout)

        layout.state = self._wdgt.saveState()





import attrs
import numpy as np
import qtawesome as qta
from qtpy import QtGui
import typing as tp

from NaviNIBS.Navigator.GUI.CollectionModels import CollectionTableModel, K, logger
from NaviNIBS.Navigator.Model.SubjectRegistration import Fiducials, Fiducial
from NaviNIBS.util import makeStrUnique

@attrs.define(slots=False)
class RegistrationFiducialsTableModel(CollectionTableModel[str, Fiducials, Fiducial]):
    _collection: Fiducials = attrs.field(init=False, repr=False)

    _checkIcon_planned: QtGui.QIcon = attrs.field(factory=lambda: qta.icon('mdi6.checkbox-marked-circle', color='blue'))
    _checkIcon_sampled: QtGui.QIcon = attrs.field(factory=lambda: qta.icon('mdi6.checkbox-marked-circle', color='green'))
    _xIcon: QtGui.QIcon = attrs.field(factory=lambda: qta.icon('mdi6.close-circle-outline', color='red'))

    def __attrs_post_init__(self):
        self._collection = self._session.subjectRegistration.fiducials

        self._attrColumns = [
            'key',
            'alignmentWeight'
        ]

        self._derivedColumns = dict(
            plannedIsSet=self._getPlannedIsSet,
            sampledIsSet=self._getSampledIsSet
        )

        self._editableColumnValidators = dict(
            key=self._checkIfKeyValid,
            alignmentWeight=self._checkIfWeightValid,
        )

        self._derivedColumnSetters = dict(
            alignmentWeight=self._setAlignmentWeight,
        )

        self._columns = [
            'key',
            'plannedIsSet',
            'sampledIsSet',
            'alignmentWeight'
        ]

        self._decoratedColumns = [
            'plannedIsSet',
            'sampledIsSet'
        ]

        self._columnLabels = dict(
            key='Fiducial',
            plannedIsSet='Planned',
            sampledIsSet='Sampled',
            alignmentWeight='Weight'
        )

        self._editableColumns.extend(['key', 'alignmentWeight'])

        self._collection.sigItemsAboutToChange.connect(self._onCollectionAboutToChange)
        self._collection.sigItemsChanged.connect(self._onCollectionChanged)

        super().__attrs_post_init__()

    def _checkIfKeyValid(self, fidKey: str, oldKey: str, newKey: str) -> bool:
        if len(newKey) == 0:
            return False

        if newKey != fidKey and newKey in self._session.subjectRegistration.fiducials:
            return False  # key is not unique

        return True

    def _getPlannedIsSet(self, fidKey: str) -> tuple[QtGui.QIcon, str]:
        if self._collection[fidKey].plannedCoord is None:
            return self._xIcon, ''
        else:
            return self._checkIcon_planned, ''

    def _getSampledIsSet(self, fidKey: str) -> tuple[QtGui.QIcon, str]:
        if self._collection[fidKey].sampledCoord is None:
            return self._xIcon, ''
        numPtsSampled = self._collection[fidKey].sampledCoords.shape[0]
        if numPtsSampled == 1:
            return self._checkIcon_sampled, ''
        else:
            return self._checkIcon_sampled, f'{numPtsSampled}'  # show an indicator of how many points have been sampled

    def _checkIfWeightValid(self, fidKey: str, oldVal: str, newVal: str) -> bool:
        if len(newVal) == 0:
            return False

        try:
            val = float(newVal)
        except Exception as e:
            return False

        if np.isnan(val) or np.isinf(val) or val < 0:
            return False

        return True

    def _setAlignmentWeight(self, fidKey: str, strVal: str) -> None:
        self._collection[fidKey].alignmentWeight = float(strVal)



# TODO: implement PlanningFiducialsTableModel for planning fiducials tab (not showing sampled fiducials at all)


@attrs.define(slots=False)
class PlanningFiducialsTableModel(CollectionTableModel[str, Fiducials, Fiducial]):
    _collection: Fiducials = attrs.field(init=False, repr=False)

    _hasPlaceholderNewRow: bool = True
    _placeholderNewRowDefaults: dict[str, tp.Any] = attrs.field(factory=lambda: dict(key='<NewFiducial>'))


    def __attrs_post_init__(self):
        self._collection = self._session.subjectRegistration.fiducials

        self._attrColumns = [
            'key'
        ]

        self._derivedColumns = dict(
            plannedCoord=self._getPlannedCoordText,
        )

        self._columns = self._attrColumns + list(self._derivedColumns.keys())

        self._columnLabels = dict(
            key='Label',
            plannedCoord='XYZ'
        )

        self._editableColumns.append('key')
        # TODO: make XYZ coord field editable too
        # (just need to implement support for editable derived fields in CollectionTableModel)

        self._collection.sigItemsAboutToChange.connect(self._onCollectionAboutToChange)
        self._collection.sigItemsChanged.connect(self._onCollectionChanged)

        self._addNewRowFromEditedPlaceholder = self.__addNewRowFromEditedPlaceholder

        super().__attrs_post_init__()

    def _getPlannedCoordText(self, fidKey: str | None) -> str:
        if fidKey is None:
            return ''
        coord = self._session.subjectRegistration.fiducials[fidKey].plannedCoord
        if coord is None:
            return ''
        else:
            return '{:.1f}, {:.1f}, {:.1f}'.format(*coord)

    def __addNewRowFromEditedPlaceholder(self, **kwargs) -> str:
        if 'key' not in kwargs:
            kwargs['key'] = makeStrUnique('Fiducial', self._collection.keys())
        item = Fiducial(**kwargs)
        self._collection.addItem(item)
        return item.key



import json
import sys

import numpy as np
import pytransform3d.rotations as ptr
import pytransform3d.transformations as ptt
from skspatial.objects import Vector
import typing as tp


def composeTransform(R: np.ndarray, p: tp.Optional[np.ndarray] = None) -> np.ndarray:
    if p is None:
        p = np.zeros((3,))
    return ptt.transform_from(R, p)


def decomposeTransform(A2B: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Extract translation and rotation components from transform. Assume the transform is rigid.
    :return: (3x3 rotation matrix, 3-elem translation vector)
    """
    R = A2B[0:3, 0:3]
    T = A2B[0:3, 3]
    return R, T


def _pttTransform_noCheck(A2B: np.ndarray, PA: np.ndarray) -> np.ndarray:
    """
    Adapted from ptt.transform but bypassing check for speed
    """
    PA = np.asarray(PA)

    if PA.ndim == 1:
        return np.dot(A2B, PA)

    if PA.ndim == 2:
        return np.dot(PA, A2B.T)

    raise ValueError("Cannot transform array with more than 2 dimensions")


def applyTransform(A2B: tp.Union[np.ndarray, tp.Iterable[np.ndarray]],
                   pts: np.ndarray,
                   doCheck: bool = True,
                   doStrictCheck: bool = True) -> np.ndarray:
    """
    Apply 4x4 transform(s) to a set of points
    :param A2B: single 4x4 transform, or iterable of 4x4 transforms. If multiple, will apply in reversed order, so that
        `applyTransform([space1ToSpace2Transf, space2TransfToSpace3Transf], pts)` correctly transforms from space1 to space3
        as might be expected with `space2TransfToSpace3Transf @ space1ToSpace2Transf @ augmentedPts`
    :param pts: Nx3 points. Or can be in shape (3,) and will return transformed points with same shape.
    :param: doCheck: whether to check that transform is valid
    :param: doStrictCheck: whether raise error if transform is not valid
    :return: transformed points
    """
    if pts.ndim == 1:
        didInsertAxis = True
        pts = pts[np.newaxis, :]
    else:
        didInsertAxis = False
        assert pts.shape[1] == 3

    if not isinstance(A2B, np.ndarray):
        A2B = concatenateTransforms(A2B)

    if not doCheck:
        # adapt ptt.transform here since it doesn't allow bypassing check
        result = _pttTransform_noCheck(A2B, ptt.vectors_to_points(pts))[:, 0:3]
    else:
        result = ptt.transform(A2B, ptt.vectors_to_points(pts), strict_check=doStrictCheck)[:, 0:3]
    if didInsertAxis:
        result = result[0, :]
    return result


def applyDirectionTransform(A2B: np.ndarray,
                            dirs: np.ndarray,
                            doCheck: bool = True,
                            doStrictCheck: bool = True) -> np.ndarray:
    if dirs.ndim == 1:
        didInsertAxis = True
        dirs = dirs[np.newaxis, :]
    else:
        didInsertAxis = False
        assert dirs.shape[1] == 3

    if not isinstance(A2B, np.ndarray):
        A2B = concatenateTransforms(A2B)

    if not doCheck:
        result = _pttTransform_noCheck(A2B, ptt.vectors_to_directions(dirs))[:, 0:3]
    else:
        result = ptt.transform(A2B, ptt.vectors_to_directions(dirs), strict_check=doStrictCheck)[:, 0:3]
    if didInsertAxis:
        result = result[0, :]
    return result


def concatenateTransforms(A2B: tp.Iterable[np.ndarray]) -> np.ndarray:
    """
    Combine transforms in **reverse** order, using same ordering convention as `applyTransform`, such that
    `applyTransform(concatenateTransforms([space1ToSpace2Transf, space2TransfToSpace3Transf]), pts)` correctly transforms from space1 to space3
    as might be expected with `space2TransfToSpace3Transf @ space1ToSpace2Transf @ augmentedPts`
    """
    A2B_combined = np.eye(4)
    for A2B_i in A2B:
        A2B_combined = A2B_i @ A2B_combined
    return A2B_combined


def invertTransform(A2B: np.ndarray) -> np.ndarray:
    return np.linalg.pinv(A2B)


def transformToString(A2B: np.ndarray, precision=12) -> str:
    return json.dumps(A2B.round(decimals=precision).tolist())


def stringToTransform(inputStr: str) -> np.ndarray:
    """
    Raises ValueError if unable to convert to valid transform
    :param inputStr: str representation of a transform, e.g. from
        `transformToString(np.asarray([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]))`
    :return: transform as 4x4 ndarray
    """
    listVal = json.loads(inputStr)
    if not isinstance(listVal, list):
        raise ValueError('Expected list of lists')
    arrayVal = np.asarray(listVal, dtype=np.float64)  # raises value error if problem
    ptt.check_transform(arrayVal)
    return arrayVal


def calculateRotationMatrixFromTwoVectors(vecA: np.ndarray, vecB: np.ndarray) -> np.ndarray:
    """
    Calculate rotation such that X-axis points in direction of vecA, Y-axis points in direction of (oproj_vecA vecB)
    adapted from https://rock-learning.github.io/pytransform3d/_apidoc/pytransform3d.rotations.matrix_from_two_vectors.html

    :param vecA:
    :param vecB:
    :return: 3x3 rotation matrix
    """
    vecA = vecA / np.linalg.norm(vecA)
    vecB = vecB / np.linalg.norm(vecB)

    assert not np.array_equal(vecA, vecB)

    dirB = vecB - vecA * np.dot(vecA, vecB)

    dirC = np.cross(vecA, dirB)

    return np.column_stack((vecA, dirB, dirC))


def calculateRotationMatrixFromVectorToVector(vecA: np.ndarray, vecB: np.ndarray) -> np.ndarray:
    """
    Calculate rotation that would rotate vecA to vecB
    """
    vecA = Vector(vecA)
    vecB = Vector(vecB)
    if vecA.is_parallel(vecB):
        return np.eye(3)
    vec_rotAxis = vecA.cross(vecB)
    rotAngle = vecA.angle_signed_3d(vecB, vec_rotAxis)  # TODO: check sign
    return ptr.matrix_from_axis_angle(np.append(vec_rotAxis, rotAngle))


def estimateAligningTransform(ptsA: np.ndarray, ptsB: np.ndarray, method: str = 'kabsch-svd', weights: tp.Optional[np.ndarray] = None) -> np.ndarray:
    """
    Estimate a transform that aligns one set of points onto another.

    Some methods assume row-wise matching of points between sets.
    
    For details of the kabsch-svd method, see:
     - https://stackoverflow.com/questions/60877274/optimal-rotation-in-3d-with-kabsch-algorithm
     - https://zpl.fi/aligning-point-patterns-with-kabsch-umeyama-algorithm/
     - http://nghiaho.com/?page_id=671
    
    :param ptsA: Nx3 ndarray of points 
    :param ptsB: Mx3 ndarray of points
    :param method: method to use for estimating transform. Default is 'kabsch-svd''
    :param weights: default None, otherwise format depends on method:
        if method == 'kabsch-svd':
            Kabsch weighted algorithm will be used. Weights should be of length equal to number of points in ptsA and ptsB.
        elif method == 'ICP':
            Weights should be of length 6, with values as defined by simpleicp's `rbp_observation_weights` argument.
    :return: A2B, 4x4 transform aligning ptsA to ptsB 
    """

    match method:
        case 'kabsch-svd':
            if ptsA.shape != ptsB.shape:
                raise ValueError('ptsA and ptsB should be matched sizes!')

            if any(pts.ndim < 2 or pts.shape[1] != 3 for pts in (ptsA, ptsB)):
                raise ValueError('pts should be of size Nx3')

            if ptsA.shape[0] < 3:
                raise ValueError('Need at least 3 points to estimate aligning transform')

            if weights is None:
                # adapted from http://nghiaho.com/?page_id=671
                centroidA = ptsA.mean(axis=0)
                centroidB = ptsB.mean(axis=0)

                ptsA_ctrd = ptsA - centroidA
                ptsB_ctrd = ptsB - centroidB

                H = ptsA_ctrd.T @ ptsB_ctrd

                U, S, Vt = np.linalg.svd(H)
                R = Vt.T @ U.T

                # reflection
                if np.linalg.det(R) < 0:
                    Vt[2, :] *= -1
                    R = Vt.T @ U.T

                t = -R @ centroidA.reshape(-1, 1) + centroidB.reshape(-1, 1)

                transf = np.eye(4)
                transf[:3, :3] = R
                transf[:3, 3] = t.reshape(3)

            else:
                from rmsd import kabsch_weighted
                [R, p, _] = kabsch_weighted(ptsB, ptsA, weights)
                transf = composeTransform(R, p)

            return transf

        case 'ICP':
            runKwargs = dict()
            if weights is not None:
                assert len(weights) == 6
                runKwargs['rbp_observation_weights'] = weights

            import simpleicp

            pc_fix = simpleicp.PointCloud(ptsB, columns=('x', 'y', 'z'))
            pc_mov = simpleicp.PointCloud(ptsA, columns=('x', 'y', 'z'))
            icp = simpleicp.SimpleICP()
            icp.add_point_clouds(pc_fix, pc_mov)
            H, X_mov_transformed, rigid_body_transformation_params, _ = icp.run(**runKwargs)

            return H

        case _:
            raise NotImplementedError()





from __future__ import annotations

import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore


def getIcon(name: str, **kwargs) -> QtGui.QIcon:
    """
    Wrapper around qta.icon but with colors that update from QPalette with each call
    instead of using qta's cached initial colors.
    """
    if True:
        return DynamicColorIconicFont.getSingleton().icon(name, **kwargs)
    else:
        kwargs.setdefault('color', qta.iconic_font.text_color())
        kwargs.setdefault('color_disabled', qta.iconic_font.text_color_disabled())
        return qta.icon(name, **kwargs)


_dynamicColorIconicFontSingleton = None


class DynamicColorIconicFont(qta.IconicFont):
    """
    Extend qta.IconicFont to automatically update to use current palette colors (if color is not manually specified)
    """

    def __init__(self, *args):
        # override init to substitute in our custom dynamic color painter
        QtCore.QObject.__init__(self)
        self.painter = DynamicColorCharIconPainter()
        self.painters = {}
        self.fontname = {}
        self.fontdata = {}
        self.fontids = {}
        self.charmap = {}
        self.icon_cache = {}
        self.rawfont_cache = {}
        for fargs in args:
            self.load_font(*fargs)

    def _parse_options(self, specific_options, general_options, name):
        # override default colors to be "None" to indicate they should be dynamically updated when painting
        override_defaults = dict(
            color=None,
            color_disabled=None
        )
        general_options = dict(override_defaults, **general_options)

        return super()._parse_options(specific_options, general_options, name)

    @classmethod
    def getSingleton(cls) -> DynamicColorIconicFont:
        global _dynamicColorIconicFontSingleton
        if _dynamicColorIconicFontSingleton is None:
            _dynamicColorIconicFontSingleton = cls(*qta._BUNDLED_FONTS)
        return _dynamicColorIconicFontSingleton


class DynamicColorCharIconPainter(qta.iconic_font.CharIconPainter):
    def _paint_icon(self, iconic, painter, rect, mode, state, options):
        options = options.copy()
        for key in ('color', 'color_on', 'color_active', 'color_selected', 'color_on_active', 'color_on_selected',
                    'color_off', 'color_off_active', 'color_off_selected'):
            if key in options and options[key] is None:
                options[key] = qta.iconic_font.text_color()
        for key in ('color_on_disabled', 'color_off_disabled'):
            if key in options and options[key] is None:
                options[key] = qta.iconic_font.text_color_disabled()
        return super()._paint_icon(iconic, painter, rect, mode, state, options)



